# Copyright (c) 2010-2017 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import

from rdkit import Chem

from . import pattern_fingerprinter
from . import rdkit
from . import types
from . import __version__ as chemfp_version

SOFTWARE = rdkit.SOFTWARE + (" chemfp/%s" % (chemfp_version,))

class HydrogenMatcher(object):
    def has_match(self, mol):
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() == 1:
                return 1
            if atom.GetTotalNumHs():
                return 1
        return 0

    def num_matches(self, mol, largest_count):
        num_hydrogens = 0
        for atom in mol.GetAtoms():
            if atom.GetAtomicNum() == 1:
                num_hydrogens += 1
            num_hydrogens += atom.GetTotalNumHs()
            if num_hydrogens >= largest_count:
                return num_hydrogens
        return num_hydrogens
        
            

class AromaticRings(object):
    def __init__(self):
        # The single ring case is easy; if there's an aromatic atom in a ring
        # then there's a ring
        self._single_matcher = Chem.MolFromSmarts("[aR]")

    def has_match(self, mol):
        return mol.HasSubstructMatch(self._single_matcher)

    def num_matches(self, mol, largest_count):
        nArom = 0
        for ring in mol.GetRingInfo().BondRings():
            if all(mol.GetBondWithIdx(bondIdx).GetIsAromatic() for bondIdx in ring):
                nArom += 1
                if nArom == largest_count:
                    return nArom
        return nArom

def _is_hetereo_aromatic_atom(atom):
    return atom.GetIsAromatic() and atom.GetAtomicNum() not in (1, 6)

class HeteroAromaticRings(object):
    def __init__(self):
        # In the single match case, if there's an aromatic non-carbon atom
        # then it's a hetereo ring
        self._single_matcher = Chem.MolFromSmarts("[aR;!#6]")

    def has_match(self, mol):
        return mol.HasSubstructMatch(self._single_matcher)

    def num_matches(self, mol, largest_count):
        nArom = 0
        for ring in mol.GetRingInfo().AtomRings():
            if any(_is_hetereo_aromatic_atom(mol.GetAtomWithIdx(atomIdx))
                                 for atomIdx in ring):
                nArom += 1
                if nArom == largest_count:
                    return nArom
        return nArom


class NumFragments(object):
    def has_match(self, mol):
        return mol.GetNumAtoms() > 0
    def num_matches(self, mol, largest_count):
        return len(Chem.GetMolFrags(mol))

# RDKit matches "molecule.HasSubstructMatch(match_pattern)"
# while every other toolkit does something like "match_pattern.HasSubstructMatch(molecule)"
# Since SMARTS doesn't handle all the pattern cases, I prefer the second ordering.
# This class inverts the order so I can do that.
class InvertedMatcher(object):
    def __init__(self, matcher):
        self.matcher = matcher
    def has_match(self, mol):
        return mol.HasSubstructMatch(self.matcher)
    def num_matches(self, mol, max_count):
        return len(mol.GetSubstructMatches(self.matcher))

_pattern_classes = {
    "<H>": HydrogenMatcher,
    "<aromatic-rings>": AromaticRings,
    "<hetero-aromatic-rings>": HeteroAromaticRings,
    "<fragments>": NumFragments,
    }
    

def rdkit_compile_pattern(pattern, max_count):
    if pattern in _pattern_classes:
        return _pattern_classes[pattern]()
    
    elif pattern.startswith("<"):
        raise NotImplementedError(pattern)
        #return NotImplemented

    # Everything else must be a SMARTS pattern

    matcher = Chem.MolFromSmarts(pattern)
    if matcher is None:
        raise pattern_fingerprinter.UnsupportedPatternError(
            pattern, "Can not interpret SMARTS pattern")
    return InvertedMatcher(matcher)

class RDKitPatternFingerprinter(pattern_fingerprinter.PatternFingerprinter):
    def __init__(self, patterns):
        assert patterns is not None
        super(RDKitPatternFingerprinter, self).__init__(patterns, rdkit_compile_pattern)
        
    def fingerprint(self, mol):
        bytes = [0] * self.num_bytes
        for matcher, largest_count, count_info_tuple in self.matcher_definitions:
            if largest_count == 1:
                if matcher.has_match(mol):
                    count_info = count_info_tuple[0]
                    bytes[count_info.byteno] |= count_info.bitmask
            else:
                actual_count = matcher.num_matches(mol, largest_count)
                if actual_count:
                    for count_info in count_info_tuple:
                        if actual_count >= count_info.count:
                            bytes[count_info.byteno] |= count_info.bitmask
                        else:
                            break
        return "".join(map(chr, bytes))

class _CachedFingerprinters(dict):
    def __missing__(self, name):
        patterns = pattern_fingerprinter._load_named_patterns(name)
        fingerprinter = RDKitPatternFingerprinter(patterns)
        self[name] = fingerprinter
        return fingerprinter
_cached_fingerprinters = _CachedFingerprinters()



_base = rdkit._base.clone(
    software = SOFTWARE)

SubstructRDKitFingerprinter_v1 = _base.clone(
    name = "ChemFP-Substruct-RDKit/1",
    num_bits = 881,
    make_fingerprinter = lambda: _cached_fingerprinters["substruct"].fingerprint)


#    def describe(self, bitno):
#        return self._fingerprinter.describe(bitno)

RDMACCSRDKitFingerprinter_v1 = _base.clone(
    name = "RDMACCS-RDKit/1",
    num_bits = 166,
    make_fingerprinter = lambda: _cached_fingerprinters["rdmaccs"].fingerprint)

RDMACCSRDKitFingerprinter_v2 = _base.clone(
    name = "RDMACCS-RDKit/2",
    num_bits = 166,
    make_fingerprinter = lambda: _cached_fingerprinters["rdmaccs2"].fingerprint)



from __future__ import absolute_import, print_function

import unittest2
from cStringIO import StringIO as BytesIO

import chemfp
from chemfp import bitops
from chemfp.bitops import hex_decode
import chemfp.search

from support import fullpath

queries_arena = chemfp.load_fingerprints(fullpath("queries.fps"))[2:]
targets_arena = chemfp.load_fingerprints(fullpath("targets.fps"))[3:-1]
MACCS_FPS = fullpath("chebi_rdmaccs.fps")

small = chemfp.load_fingerprints(BytesIO(b"""\
012345\tfirst
01ABCD\tsecond
672245\tthird
"""))

class ContainsFPSearch(unittest2.TestCase):
    def test_contains_fp_zero_query(self):
        hits = chemfp.search.contains_fp(hex_decode("000000"), small)
        self.assertEqual(len(hits), 3)

    def test_contains_fp_find_some(self):
        hits = chemfp.search.contains_fp(hex_decode("010300"), small)
        self.assertEqual(len(hits), 2)

    def test_contains_fp_find_none(self):
        hits = chemfp.search.contains_fp(hex_decode("67AB00"), small)
        self.assertEqual(len(hits), 0)

    def test_alignment(self):
        left_bits = hex_decode("010000000000000000000000000000")
        right_bits = hex_decode("000000000000000000000000010000")
        no_bits = hex_decode("000000000000000000000000010001")
        for alignment in (1, 4, 8, 32):
            arena = chemfp.load_fingerprints(BytesIO(b"010000000100000001000000010000\tfirst\n"),
                                             alignment=alignment)
            hits = chemfp.search.contains_fp(left_bits, arena)
            assert len(hits) == 1
            hits = chemfp.search.contains_fp(right_bits, arena)
            assert len(hits) == 1
            hits = chemfp.search.contains_fp(no_bits, arena)
            assert len(hits) == 0

    def test_single_bit(self):
        template = b"\0" * 11
        id_fp_pairs = [(None, template)]
        for i in range(len(template)*8):
            offset = i//8
            byte = b"\x01\x02\x04\x08\x10\x20\x40\x80"[i%8:i%8+1]
            fp = template[:offset] + byte + template[offset+1:]
            id_fp_pairs.append((i, fp))

        for alignment in (1, 2, 4, 8, 16):
            arena = chemfp.load_fingerprints(id_fp_pairs,
                                             metadata=chemfp.Metadata(num_bytes=len(template)),
                                             alignment=alignment)
            self.assertEqual(arena.alignment, alignment)
            first_fp = arena.get_fingerprint(0)
            self.assertNotEqual(arena.get_fingerprint(1)[0:1], b"\0")
            self.assertEqual(arena.get_fingerprint(1)[-1:], b"\0")
            self.assertNotEqual(arena.popcount_indices, "")

            for query_id, query_fp in id_fp_pairs[1:]:
                hits = chemfp.search.contains_fp(query_fp, arena)
                self.assertEqual(len(hits), 1)
                self.assertEqual(hits.get_ids(), [query_id])

            # And try it in reverse order, without the empty fingerprint
            arena = chemfp.load_fingerprints(id_fp_pairs[:0:-1],
                                             metadata=chemfp.Metadata(num_bytes=len(template)),
                                             alignment=alignment)
            self.assertEqual(arena.alignment, alignment)
            self.assertNotEqual(arena.popcount_indices, "")
            first_fp = arena.get_fingerprint(0)
            self.assertEqual(arena.get_fingerprint(0)[0:1], b"\0")
            self.assertNotEqual(arena.get_fingerprint(0)[-1:], b"\0")

            for query_id, query_fp in id_fp_pairs[1:]:
                hits = chemfp.search.contains_fp(query_fp, arena)
                self.assertEqual(len(hits), 1)
                self.assertEqual(hits.get_ids(), [query_id])

            # And try it without popcounts.
            # Use reverse ordering to ensure no popcounts are set.
            arena = chemfp.load_fingerprints(id_fp_pairs[::-1],
                                             metadata=chemfp.Metadata(num_bytes=len(template)),
                                             alignment=alignment, reorder=False)
            self.assertEqual(arena.alignment, alignment)
            self.assertEqual(arena.popcount_indices, b"")
            first_fp = arena.get_fingerprint(0)
            self.assertEqual(arena.get_fingerprint(0)[0:1], b"\0")
            self.assertNotEqual(arena.get_fingerprint(0)[-1:], b"\0")

            for query_id, query_fp in id_fp_pairs[1:]:
                hits = chemfp.search.contains_fp(query_fp, arena)
                self.assertEqual(len(hits), 1)
                self.assertEqual(hits.get_ids(), [query_id])
                 
            
class TestCrossComparison(unittest2.TestCase):
    def test_1024_bit_fingerprints(self):
        target_fingerprints = [fp for id, fp in targets_arena]

        row_hits = []
        for query_id, query_fp in queries_arena:
            hits = chemfp.search.contains_fp(query_fp, targets_arena)
            expected = sum(1 for target_fp in target_fingerprints if bitops.byte_contains(query_fp, target_fp))
            self.assertEqual(len(hits), expected)
            row_hits.append(hits)

        for hits, expected_hits in zip(chemfp.search.contains_arena(queries_arena, targets_arena), row_hits):
            self.assertEqual(len(hits), len(expected_hits))

    def test_166_bit_fingerprints(self):
        for alignment in (1, 4, 8):
            arena = chemfp.load_fingerprints(MACCS_FPS, alignment=alignment)[100:169]
            fingerprints = [fp for id, fp in arena]

            row_hits = []
            for query_id, query_fp in arena:
                hits = chemfp.search.contains_fp(query_fp, arena)
                expected = sum(1 for target_fp in fingerprints if bitops.byte_contains(query_fp, target_fp))
                self.assertEqual(len(hits), expected)
                row_hits.append(hits)

            for hits, expected_hits in zip(chemfp.search.contains_arena(arena, arena), row_hits):
                self.assertEqual(len(hits), len(expected_hits))

    def test_166_bit_fingerprints_with_no_popcount_index(self):
        # Excercise the code path which doesn't use the popcount index
        for alignment in (1, 4, 8):
            arena = chemfp.load_fingerprints(MACCS_FPS, alignment=alignment, reorder=False)[100:169]
            self.assertFalse(arena.popcount_indices)
            fingerprints = [fp for id, fp in arena]

            row_hits = []
            for query_id, query_fp in arena:
                hits = chemfp.search.contains_fp(query_fp, arena)
                expected = sum(1 for target_fp in fingerprints if bitops.byte_contains(query_fp, target_fp))
                self.assertEqual(len(hits), expected)
                row_hits.append(hits)

            for hits, expected_hits in zip(chemfp.search.contains_arena(arena, arena), row_hits):
                self.assertEqual(len(hits), len(expected_hits))

            
if __name__ == "__main__":
    unittest2.main()



# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import, print_function
import os
import sys
import itertools
import signal

from .. import argparse
from .. import __version__ as chemfp_version
from .. import ChemFPError, ParseError
from .. import io

try:
    import chemfp_converters
except ImportError:
    has_chemfp_converters = False
else:
    has_chemfp_converters = True


######
# Code extracted from Python's argparse.py. Distributed under the Python license.
# Before Python 3.4, argparse --version sent the message to stderr, not stdout.
# The bug was fixed in https://bugs.python.org/issue18920 .
# chemfp supports Python 2.7 and Python 3.5+.
# This is a back-port to version 2.7

version_action = "version"
#if sys.version_info.major == 2:
if sys.version_info[0] == 2: # Support Python 2.6
    class ChemfpVersionAction(argparse._VersionAction):
        def __call__(self, parser, namespace, values, option_string=None):
            version = self.version
            if version is None:
                version = parser.version
            formatter = parser._get_formatter()
            formatter.add_text(version)
            parser._print_message(formatter.format_help(), sys.stdout)
            parser.exit()

    version_action = ChemfpVersionAction
    
def add_version(parser):
    parser.add_argument("--version", action=version_action,
                        version="%(prog)s " + chemfp_version)
    

######

def mutual_exclusion(parser, args, default, groups):
    true_groups = []
    for g in groups:
        if getattr(args, g):
            true_groups.append(g)

    if not true_groups:
        setattr(args, default, True)
    elif len(true_groups) == 1:
        pass
    else:
        parser.error("Cannot specify both --%s and --%s" % (true_groups[0], true_groups[1]))

class AbortHandler(object):
    def error(self, msg, location):
        raise ParseError("Each of the first 100 records contained errors. Final error: %s" % (msg,), location)

def cleanup_ids(reader, id_tag, error_handler, location):
    err_count = 0
    for recno, (id, fp) in enumerate(reader):
        if not id:
            if recno == err_count == 99:
                error_handler = AbortHandler()
            err_count += 1
                
            if location.record_format == "sdf":
                if id_tag is None:
                    error_handler.error("Missing title in SD record", location)
                else:
                    error_handler.error("Missing id tag %r in SD record" % (id_tag,), location)
            elif location.record_format == "smi":
                error_handler.error("Missing SMILES identifier (second column)", location)
            else:
                error_handler.error("Missing identifier", location)
            continue
                    
        else:
            # This assumes that these characters are rare
            if "\n" in id:
                id = id.partition("\n")[0]  # get up to the newline
            if "\r" in id:
                id = id.replace("\r", "")
            if "\t" in id:
                id = id.replace("\t", "")
            if "\0" in id:
                id = id.replace("\0", "")
            if " " in id:
                id = id.strip()

            if not id:
                if recno == err_count == 99:
                    error_handler = AbortHandler()
                err_count += 1
                if location.record_format == "sdf":
                    if id_tag is None:
                        error_handler.error("Empty title in SD record after cleanup", location)
                    else:
                        error_handler.error("Empty id tag %r in SD record after cleanup"
                                            % (id_tag,), location)
                elif location.record_format == "smi":
                    error_handler.error("Empty SMILES identifier (second column) after cleanup", location)
                else:
                    error_handler.error("Empty identifier after cleanup", location)
                continue
            else:
                yield id, fp
    

def sys_exit_opener(opener, source, format, id_tag, error_handler, reader_args, location):
    try:
        return opener.read_molecule_fingerprints(source, format, id_tag, reader_args=reader_args,
                                                 errors=error_handler, location=location)
    except (IOError, ChemFPError, ValueError) as err:
        sys.stderr.write("Problem reading structure fingerprints: %s. Exiting.\n" % err)
        raise SystemExit(1)

## def iter_all_sources(fptype, filenames, format, id_tag, errors):
##     for filename in filenames:
##         reader = sys_exit_opener(fptype, filename, format, id_tag, errors)
##         for id, fp in reader:
##             yield id, fp

def read_multifile_structure_fingerprints(opener, filenames, format, id_tag, reader_args, errors):
    error_handler = io.get_parse_error_handler(errors)
    if errors == "strict":
        id_error_handler = error_handler
    else:
        id_error_handler = io.get_parse_error_handler("report")
    
    if not filenames:
        location = io.Location.from_source(None)
        reader = sys_exit_opener(opener, None, format, id_tag, error_handler, reader_args, location)
        cleanup_reader = cleanup_ids(reader, id_tag, id_error_handler, location)
        return reader.metadata, cleanup_reader

    if len(filenames) == 1:
        location = io.Location.from_source(filenames[0])
        reader = sys_exit_opener(opener, filenames[0], format, id_tag, error_handler, reader_args, location)
        cleanup_reader = cleanup_ids(reader, id_tag, id_error_handler, location)
        return reader.metadata, cleanup_reader

    location = io.Location.from_source(filenames[0])
    reader = sys_exit_opener(opener, filenames[0], format, id_tag, error_handler, reader_args, location)
    reader.metadata.sources = filenames

    def multifile_reader(reader):
        # This is very annoying. The gzip reader fails if the file
        # isn't in the correct format but the check doesn't occur
        # until someone tries to read a byte. The error message is the
        # un-helpful generic "Not a gzipped file", without the
        # filename. I want to report that error, but do so carefully.

        
        try:
            # Handle the first filename
            filename = filenames[0]
            
            for id, fp in cleanup_ids(reader, id_tag, id_error_handler, location):
                yield id, fp
        
            # Handle the remaining filenames
            for filename in filenames[1:]:
                location.filename = filename
                reader = sys_exit_opener(opener, filename, format, id_tag,
                                         error_handler, reader_args, location)
                for id, fp in cleanup_ids(reader, id_tag, id_error_handler, location):
                    yield id, fp

        except IOError as err:
            if err.filename is None:
                raise IOError("%s: %r" % (err, filename))
            raise
        except EOFError as err:
            raise EOFError("%s: %r" % (err, filename))
            

    return reader.metadata, multifile_reader(reader)

def is_valid_tag(tag):
    if tag is None:
        return True
    for c in "<>\r\n":
        if c in tag:
            return False
    return True

def check_filenames(parser, toolkit, format, filenames):
    if not filenames:
        # Read from stdin.
        return None
    for filename in filenames:
        if not os.path.exists(filename):
            parser.error("Structure file %r does not exist" % (filename,))
        ## try:
        ##     toolkit.get_input_format_from_source(source=filename, format=format)
        ## except ValueError as err:
        ##     parser.error(str(err))

def run(main_func):
    signal.signal(signal.SIGINT, signal.SIG_DFL)  # Allow ^C to kill the process.
    signal.signal(signal.SIGPIPE, signal.SIG_DFL) # Allow the output pipe to be closed
    try:
        main_func()
    except KeyboardInterrupt:
        raise SystemExit(2)

def die(msg):
    sys.stderr.write(msg)
    sys.stderr.write("\n")
    sys.stderr.flush()
    raise SystemExit(1)



"""Algorithms and data structure for working with a FingerprintArena.

NOTE: This module should not be used directly.

A FingerprintArena stores the fingerprints as a contiguous byte
string, called the `arena`. Each fingerprint takes `storage_size`
bytes, which may be larger than `num_bytes` if the fingerprints have a
specific memory alignment. The bytes for fingerprint i are
  arena[i*storage_size:i*storage_size+num_bytes]
Additional bytes must contain NUL bytes.

The lookup for `ids[i]` contains the id for fingerprint `i`.

A FingerprintArena has an optional `indices` attribute. When
available, it means that the arena fingerprints and corresponding ids
are ordered by population count, and the fingerprints with popcount
`p` start at index indices[p] and end just before indices[p+1].

"""

# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import

import ctypes
from cStringIO import StringIO
import array
import itertools

from chemfp import FingerprintReader
import _chemfp
from chemfp import bitops, search
from chemfp._compat import next, BytesIO, xrange, tobytes

__all__ = []
    
class FingerprintArena(FingerprintReader):
    """Store fingerprints in a contiguous block of memory for fast searches

    A fingerprint arena implements the :class:`chemfp.FingerprintReader` API.

    A fingerprint arena stores all of the fingerprints in a continuous
    block of memory, so the per-molecule overhead is very low.

    The fingerprints can be sorted by popcount, so the fingerprints
    with no bits set come first, followed by those with 1 bit, etc.
    If ``self.popcount_indices`` is a non-empty string then the string
    contains information about the start and end offsets for all the
    fingerprints with a given popcount. This information is used for
    the sublinear search methods.

    The public attributes are:

    .. py:attribute:: metadata

       :class:`chemfp.Metadata` about the fingerprints

    .. py:attribute:: ids

       list of identifiers, in index order

    Other attributes, which might be subject to change, and which I won't fully explain, are:
      * arena - a contiguous block of memory, which contains the fingerprints
      * start_padding - number of bytes to the first fingerprint in the block
      * end_padding - number of bytes after the last fingerprint in the block
      * storage_size - number of bytes used to store a fingerprint
      * num_bytes - number of bytes in each fingerprint (must be <= storage_size)
      * num_bits - number of bits in each fingerprint
      * alignment - the fingerprint alignment
      * start - the index for the first fingerprint in the arena/subarena
      * end - the index for the last fingerprint in the arena/subarena
      * arena_ids - all of the identifiers for the parent arena

    The FingerprintArena is its own context manager, but it does
    nothing on context exit.
    """
    _search = search
    def __init__(self, metadata, alignment,
                 start_padding, end_padding, storage_size, arena,
                 popcount_indices, arena_ids, start=0, end=None,
                 id_lookup=None, num_bits=None, num_bytes=None,
                 ):
        assert isinstance(popcount_indices, bytes), type(popcount_indices) # XXX REMOVE ME
        if num_bits is None:
            num_bits = metadata.num_bits
            if num_bits is None:
                raise TypeError("Missing 'num_bits' and not available from the metadata")
        if num_bytes is None:
            num_bytes = metadata.num_bytes
            if num_bytes is None:
                raise TypeError("Missing 'num_bytes' and not available from the metadata")

        self.metadata = metadata
        self.alignment = alignment
        self.num_bits = num_bits
        self.num_bytes = num_bytes
        self.start_padding = start_padding
        self.end_padding = end_padding
        self.storage_size = storage_size
        self.arena = arena
        self.popcount_indices = popcount_indices
        self.arena_ids = arena_ids
        self.start = start   # the starting index in the arena (not byte position!)
        if end is None:      # the ending index in the arena (not byte position!)
            if num_bytes:
                end = (len(arena) - start_padding - end_padding) // self.storage_size
            else:
                end = 0
        self.end = end
        if self.start == 0 and self.end == len(arena_ids):
            self._ids = arena_ids
        else:
            self._ids = None
        self._id_lookup = id_lookup
        assert end >= start
        self._range_check = xrange(end-start)

    def __len__(self):
        """Number of fingerprint records in the FingerprintArena"""
        return self.end - self.start

    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        return  # XXX This should close the memory map
    

    @property
    def ids(self):
        """The identifiers in this arena or subarena"""
        ids = self._ids
        if ids is None:
            ids = self.arena_ids[self.start:self.end]
            self._ids = ids
        return ids

    def __getitem__(self, i):
        """Return the (id, fingerprint) pair at index i"""
        if isinstance(i, slice):
            start, end, step = i.indices(self.end - self.start)
            if step != 1:
                raise IndexError("arena slice step size must be 1")
            if start >= end:
                return FingerprintArena(self.metadata, self.alignment,
                                        0, 0, self.storage_size, b"",
                                        b"", [], 0, 0)
            return FingerprintArena(self.metadata, self.alignment,
                                    self.start_padding, self.end_padding,
                                    self.storage_size, self.arena,
                                    self.popcount_indices, self.arena_ids,
                                    self.start+start, self.start+end,
                                    self._id_lookup, self.num_bits, self.num_bytes)
        try:
            i = self._range_check[i]
        except IndexError:
            raise IndexError("arena fingerprint index out of range")
        arena_i = i + self.start
        start_offset = arena_i * self.storage_size + self.start_padding
        end_offset = start_offset + self.num_bytes
        return self.arena_ids[arena_i], self.arena[start_offset:end_offset]

    def get_fingerprint(self, i):
        """Return the fingerprint at index *i*

        Raises an IndexError if index *i* is out of range.
        """
        try:
            i = self._range_check[i]
        except IndexError:
            raise IndexError("arena fingerprint index out of range")
        arena_i = i + self.start
        start_offset = arena_i * self.storage_size + self.start_padding
        end_offset = start_offset + self.num_bytes
        return self.arena[start_offset:end_offset]

    def _make_id_lookup(self):
        d = dict((id, i) for (i, id) in enumerate(self.ids))
        self._id_lookup = d.get
        return self._id_lookup
        
    def get_by_id(self, id):
        """Given the record identifier, return the (id, fingerprint) pair,

        If the *id* is not present then return None.
        """
        id_lookup = self._id_lookup
        if id_lookup is None:
            id_lookup = self._make_id_lookup()
        i = id_lookup(id)
        if i is None:
            return None
        arena_i = i + self.start
        start_offset = arena_i * self.storage_size + self.start_padding
        end_offset = start_offset + self.num_bytes
        return self.arena_ids[arena_i], self.arena[start_offset:end_offset]

    def get_index_by_id(self, id):
        """Given the record identifier, return the record index

        If the *id* is not present then return None.
        """
        id_lookup = self._id_lookup
        if id_lookup is None:
            id_lookup = self._make_id_lookup()
        return id_lookup(id)

    def get_fingerprint_by_id(self, id):
        """Given the record identifier, return its fingerprint

        If the *id* is not present then return None
        """
        id_lookup = self._id_lookup
        if id_lookup is None:
            id_lookup = self._make_id_lookup()
        i = id_lookup(id)
        if i is None:
            return None
        arena_i = i + self.start
        start_offset = arena_i * self.storage_size + self.start_padding
        end_offset = start_offset + self.num_bytes
        return self.arena[start_offset:end_offset]

    def save(self, destination, format=None):
        """Save the arena contents to the given filename or file object"""
        from . import io
        format_name, compression = io.normalize_output_format(destination, format,
                                                              default = ("fps", ""))

        if format_name == "fps":
            output, close = io.open_binary_output(destination)
            
            try:
                io.write_fps1_magic(output)
                io.write_fps1_header(output, self.metadata)
                try:
                    for i, (id, fp) in enumerate(self):
                        io.write_fps1_fingerprint(output, fp, id)
                except ValueError as err:
                    raise ValueError("%s in record %i" % (err, i+1))
            finally:
                if close is not None:
                    close()

        elif format_name == "fpb":
            raise NotImplementedError("fpb format not implemented")
        elif format_name == "flush":
            try:
                from chemfp_converters import flush
            except ImportError:
                raise ValueError("Cannot write to flush files because the chemfp_converter module is not available")
            if compression:
                raise ValueError("Compression of flush files is not supported")
            with flush.open_fingerprint_writer(destination, metadata=self.metadata) as writer:
                writer.write_fingerprints(self)
        else:
            raise ValueError("Unknown output format %r" % (format_name,))
    save.__doc__ = FingerprintReader.save.__doc__

                
    def __iter__(self):
        """Iterate over the (id, fingerprint) contents of the arena"""
        storage_size = self.storage_size
        if not storage_size:
            return
        target_fp_size = self.num_bytes
        arena = self.arena
        start_padding = self.start_padding
        for i in xrange(self.start, self.end):
            arena_start = i*storage_size+start_padding
            yield self.arena_ids[i], arena[arena_start:arena_start+target_fp_size]

    def iter_arenas(self, arena_size = 1000):
        """Iterate through *arena_size* fingerprints at a time, as subarenas"""
        if arena_size is None:
            yield self
            return
        
        storage_size = self.storage_size
        start = self.start
        for i in xrange(0, len(self), arena_size):
            end = start+arena_size
            if end > self.end:
                end = self.end
            yield FingerprintArena(self.metadata, self.alignment,
                                   self.start_padding, self.end_padding,
                                   storage_size, self.arena,
                                   self.popcount_indices, self.arena_ids, start, end)
            start = end

    iter_arenas.__doc__ = FingerprintReader.__doc__

    def count_tanimoto_hits_fp(self, query_fp, threshold=0.7):
        """Count the fingerprints which are sufficiently similar to the query fingerprint

        Return the number of fingerprints in the arena which are
        at least *threshold* similar to the query fingerprint *query_fp*.
        
        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: integer count
        """
        return search.count_tanimoto_hits_fp(query_fp, self, threshold)

    def count_tanimoto_hits_arena(self, queries, threshold=0.7):
        """Count the fingerprints which are sufficiently similar to each query fingerprint

        DEPRECATED: Use `chemfp.search.count_tanimoto_hits_arena`_ or
        `chemfp.search.count_tanimoto_hits_symmetric`_ instead.
        
        Returns a list containing a count for each query fingerprint
        in the *queries* arena. The count is the number of
        fingerprints in the arena which are at least *threshold*
        similar to the query fingerprint.

        The order of results is the same as the order of the queries.
        
        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: list of integer counts, one for each query
        """
        return search.count_tanimoto_hits_arena(queries, self, threshold)

    def threshold_tanimoto_search_fp(self, query_fp, threshold=0.7):
        """Find the fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this arena which are at least
        *threshold* similar to the query fingerprint *query_fp*.  The
        hits are returned as a :class:`.SearchResult`, in arbitrary
        order.
        
        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        return search.threshold_tanimoto_search_fp(query_fp, self, threshold)

    def threshold_tanimoto_search_arena(self, queries, threshold=0.7):
        """Find the fingerprints which are sufficiently similar to each of the query fingerprints

        DEPRECATED: Use `chemfp.search.threshold_tanimoto_search_arena`_
        or `chemfp.search.threshold_tanimoto_search_symmetric`_ instead.

        For each fingerprint in the *queries* arena, find all of the
        fingerprints in this arena which are at least *threshold*
        similar. The hits are returned as a :class:`.SearchResults`,
        where the hits in each :class:`.SearchResult` is in arbitrary
        order.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResults`
        """
        return search.threshold_tanimoto_search_arena(queries, self, threshold)

    def knearest_tanimoto_search_fp(self, query_fp, k=3, threshold=0.7):
        """Find the k-nearest fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this arena which are at least
        *threshold* similar to the query fingerprint, and of those, select
        the top *k* hits. The hits are returned as a :class:`.SearchResult`,
        sorted from highest score to lowest.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        return search.knearest_tanimoto_search_fp(query_fp, self, k, threshold)

    def knearest_tanimoto_search_arena(self, queries, k=3, threshold=0.7):
        """Find the k-nearest fingerprints which are sufficiently similar to each of the query fingerprints

        DEPRECATED: Use `chemfp.search.knearest_tanimoto_search_arena`_ or
        `chemfp.search.knearest_tanimoto_search_symmetric`_ instead.

        For each fingerprint in the *queries* arena, find the
        fingerprints in this arena which are at least *threshold*
        similar to the query fingerprint, and of those, select the top
        *k* hits. The hits are returned as a :class:`.SearchResults`,
        where the hits in each :class:`.SearchResult` are sorted by
        similarity score.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResults`
        """
        return search.knearest_tanimoto_search_arena(queries, self, k, threshold)
    
    def count_tversky_hits_fp(self, query_fp, threshold=0.7, alpha=1.0, beta=1.0):
        """Count the fingerprints which are sufficiently similar to the query fingerprint

        Return the number of fingerprints in the arena which are
        at least *threshold* similar to the query fingerprint *query_fp*.
        
        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: integer count
        """
        return search.count_tversky_hits_fp(query_fp, self, threshold, alpha, beta)

    def threshold_tversky_search_fp(self, query_fp, threshold=0.7, alpha=1.0, beta=1.0):
        """Find the fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this arena which are at least
        *threshold* similar to the query fingerprint *query_fp*.  The
        hits are returned as a :class:`.SearchResult`, in arbitrary
        order.
        
        :param query_fp: query fingerprint
        :type query_fp: byte string
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        return search.threshold_tversky_search_fp(query_fp, self, threshold, alpha, beta)

    def knearest_tversky_search_fp(self, query_fp, k=3, threshold=0.7, alpha=1.0, beta=1.0):
        """Find the k-nearest fingerprints which are sufficiently similar to the query fingerprint

        Find all of the fingerprints in this arena which are at least
        *threshold* similar to the query fingerprint, and of those, select
        the top *k* hits. The hits are returned as a :class:`.SearchResult`,
        sorted from highest score to lowest.

        :param queries: query fingerprints
        :type queries: a :class:`.FingerprintArena`
        :param threshold: minimum similarity threshold (default: 0.7)
        :type threshold: float between 0.0 and 1.0, inclusive
        :returns: a :class:`.SearchResult`
        """
        return search.knearest_tversky_search_fp(query_fp, self, k, threshold, alpha, beta)

    
    def copy(self, indices=None, reorder=None):
        """Create a new arena using either all or some of the fingerprints in this arena

        By default this create a new arena. The fingerprint data block and ids may
        be shared with the original arena, which makes this a shallow copy. If the
        original arena is a slice, or "sub-arena" of an arena, then the copy will
        allocate new space to store just the fingerprints in the slice and use its
        own list for the ids.

        The *indices* parameter, if not None, is an iterable which contains the
        indicies of the fingerprint records to copy. Duplicates are allowed, though
        discouraged.

        If *indices* are specified then the default *reorder* value of None, or
        the value True, will reorder the fingerprints for the new arena by popcount.
        This improves overall search performance. If *reorder* is False then the
        new arena will preserve the order given by the indices.

        If *indices* are not specified, then the default is to preserve the order
        type of the original arena. Use ``reorder=True`` to always reorder the
        fingerprints in the new arena by popcount, and ``reorder=False`` to always
        leave them in the current ordering.

            >>> import chemfp
            >>> arena = chemfp.load_fingerprints("pubchem_queries.fps")
            >>> arena.ids[1], arena.ids[5], arena.ids[10], arena.ids[18]
            (b'9425031', b'9425015', b'9425040', b'9425033')
            >>> len(arena)
            19
            >>> new_arena = arena.copy(indices=[1, 5, 10, 18])
            >>> len(new_arena)
            4
            >>> new_arena.ids
            [b'9425031', b'9425015', b'9425040', b'9425033']
            >>> new_arena = arena.copy(indices=[18, 10, 5, 1], reorder=False)
            >>> new_arena.ids
            [b'9425033', b'9425040', b'9425015', b'9425031']

        :param indices: indicies of the records to copy into the new arena
        :type indices: iterable containing integers, or None
        :param reorder: describes how to order the fingerprints
        :type reorder: True to reorder, False to leave in input order, None for default action
        """
        if reorder is None:
            if indices is None:
                # This is a pure copy. Reorder only if there are popcount indices.
                reorder = (self.popcount_indices != b"")
            else:
                # The default is to go fast. If you want to preserve index order
                # then you'll need to set reorder=False
                reorder = True
            
        if indices is None:
            # Make a completely new arena
            # Handle the trivial case where I don't need to do anything.
            if (self.start == 0 and
                (self.end*self.storage_size + self.start_padding + self.end_padding == len(self.arena)) and
                (not reorder or self.popcount_indices)):
                return FingerprintArena(self.metadata, self.alignment,
                                        self.start_padding, self.end_padding, self.storage_size, self.arena,
                                        self.popcount_indices, self.arena_ids,
                                        start = 0, end = self.end,
                                        id_lookup = self._id_lookup)
            
            # Otherwise I need to do some work
            # Make a copy of the actual fingerprints. (Which could be a subarena.)
            start = self.start_padding + self.start*self.storage_size
            end = self.start_padding + self.end*self.storage_size
            arena = self.arena[start:end]

            # If we don't have popcount_indices and don't want them ordered
            # then just do the alignment and we're done.
            if not reorder and not self.popcount_indices:
                # Don't reorder the unordered fingerprints
                start_padding, end_padding, unsorted_arena = (
                    _chemfp.make_unsorted_aligned_arena(arena, self.alignment))
                return FingerprintArena(self.metadata, self.alignment, start_padding, end_padding,
                                        self.storage_size, unsorted_arena, b"", self.ids,
                                        id_lookup = self._id_lookup)

            # Either we're already sorted or we should become sorted.
            # If we're sorted then make_sorted_aligned_arena will detect
            # that and keep the old arena. Otherwise it sorts first and
            # makes a new arena block.
            current_ids = self.ids
            ordering = (ChemFPOrderedPopcount*len(current_ids))()
            popcounts = array.array("i", (0,)*(self.metadata.num_bits+2))
            start_padding, end_padding, arena = _chemfp.make_sorted_aligned_arena(
                self.metadata.num_bits, self.storage_size, arena, len(current_ids),
                ordering, popcounts, self.alignment)

            reordered_ids = [current_ids[item.index] for item in ordering]
            return FingerprintArena(self.metadata, self.alignment,
                                    start_padding, end_padding, self.storage_size,
                                    arena, popcounts.tostring(), reordered_ids)

        # On this pathway, we want to make a new arena which contains
        # selected fingerprints given indices into the old arena.
        
        arena = self.arena
        storage_size = self.storage_size
        start = self.start
        start_padding = self.start_padding
        arena_ids = self.arena_ids
        
        # First make sure that all of the indices are in range.
        # This will also convert negative indices into positive ones.
        new_indices = []
        range_check = self._range_check
        try:
            for i in indices:
                new_indices.append(range_check[i])
        except IndexError:
            raise IndexError("arena fingerprint index %d is out of range" % (i,))

        if reorder and self.popcount_indices:
            # There's a slight performance benefit because
            # make_sorted_aligned_arena will see that the fingerprints
            # are already in sorted order and not resort.
            # XXX Is that true? Why do a Python sort instead of a C sort?
            # Perhaps because then I don't need to copy fingerprints?
            new_indices.sort()

        # Copy the fingerprints over to a new arena block
        unsorted_fps = []
        new_ids = []
        for new_i in new_indices:
            start_offset = start_padding + new_i*storage_size
            end_offset = start_offset + storage_size
            unsorted_fps.append(arena[start_offset:end_offset])
            new_ids.append(arena_ids[new_i])
                
        unsorted_arena = b"".join(unsorted_fps)
        unsorted_fps = None   # regain some memory

        # If the caller doesn't want ordered data, then leave it unsorted
        if not reorder:
            start_padding, end_padding, unsorted_arena = _chemfp.make_unsorted_aligned_arena(
                unsorted_arena, self.alignment)
            return FingerprintArena(self.metadata, self.alignment, start_padding, end_padding, storage_size,
                                    unsorted_arena, b"", new_ids)

        # Otherwise, reorder and align the area, along with popcount information
        ordering = (ChemFPOrderedPopcount*len(new_ids))()
        popcounts = array.array("i", (0,)*(self.metadata.num_bits+2))

        start_padding, end_padding, sorted_arena = _chemfp.make_sorted_aligned_arena(
            self.metadata.num_bits, storage_size, unsorted_arena, len(new_ids),
            ordering, popcounts, self.alignment)

        reordered_ids = [new_ids[item.index] for item in ordering]
        return FingerprintArena(self.metadata, self.alignment,
                                start_padding, end_padding, storage_size,
                                sorted_arena, popcounts.tostring(), reordered_ids)
        

# TODO: push more of this malloc-management down into C
class ChemFPOrderedPopcount(ctypes.Structure):
    _fields_ = [("popcount", ctypes.c_int),
                ("index", ctypes.c_int)]


_methods = bitops.get_methods()
_has_popcnt = "POPCNT" in _methods
_has_ssse3 = "ssse3" in _methods

def get_optimal_alignment(num_bits):
    if num_bits <= 32:
        # Just in case!
        if num_bits <= 8:
            return 1
        return 4

    # Since the ssse3 method must examine at least 512 bits while the
    # Gillies method doesn't, this puts the time tradeoff around 210 bits.
    # I decided to save a bit of space and round that up to 224 bits.
    # (Experience will tell us if 256 is a better boundary.)
    if num_bits <= 224:
        return 8

    # If you have POPCNT (and you're using it) then there's no reason
    # to use a larger alignment
    if _has_popcnt:
        if num_bits >= 768:
            if bitops.get_alignment_method("align8-large") == "POPCNT":
                return 8
        else:
            if bitops.get_alignment_method("align8-small") == "POPCNT":
                return 8

    # If you don't have SSSE3 or you aren't using it, then use 8
    if not _has_ssse3 or bitops.get_alignment_method("align-ssse3") != "ssse3":
        return 8

    # In my timing tests:
    #    Lauradoux takes 12.6s
    #    ssse3 takes in 9.0s
    #    Gillies takes 22s


    # Otherwise, go ahead and pad up to 64 bytes
    # (Even at 768 bits/96 bytes, the SSSE3 method is faster.)
    return 64

def _get_num_bits_and_bytes(fps_reader, metadata):
    num_bits = metadata.num_bits
    num_bytes = metadata.num_bytes
    
    fps_reader_iter = iter(fps_reader)

    num_bytes_source = "the metadata"
    if num_bits is None:
        if num_bytes is None:
            # Grr. Okay, we can look-ahead by one to get the content
            try:
                id, fp = next(fps_reader_iter)
            except StopIteration:
                # No size, no fingerprints? No problem!
                num_bits = num_bytes = 0
                fps_reader_iter = iter([])
                num_bytes_source = "the lack of metadata size or fingerprints"
            else:
                # Ha! Got the number of bytes
                num_bytes = len(fp)
                num_bits = num_bytes*8
                fps_reader_iter = itertools.chain([(id, fp)], fps_reader_iter)
                num_bytes_source = "the first fingerprint"
        else:
            num_bits = num_bytes*8
    else:
        if num_bytes is None:
            num_bytes = (num_bits+7)//8
        else:
            # Check for compatibility here?
            pass
    return num_bits, num_bytes, num_bytes_source, fps_reader_iter

def fps_to_arena(fps_reader, metadata=None, reorder=True, alignment=None):
    if metadata is None:
        metadata = fps_reader.metadata
            
    num_bits, num_bytes, num_bytes_source, fps_reader_iter = _get_num_bits_and_bytes(fps_reader, metadata)

    if alignment is None:
        alignment = get_optimal_alignment(num_bits)

    storage_size = num_bytes
    if storage_size % alignment != 0:
        n = alignment - storage_size % alignment
        end_padding = "\0" * n
        storage_size += n
    else:
        end_padding = None

    ids = []
    unsorted_fps = StringIO()
    for (id, fp) in fps_reader_iter:
        if len(fp) != num_bytes:
            raise ValueError("Fingerprint for id %r has %d bytes while %s says it should have %d"
                             % (id, len(fp), num_bytes_source, num_bytes))
        unsorted_fps.write(fp)
        if end_padding:
            unsorted_fps.write(end_padding)
        ids.append(id)

    unsorted_arena = unsorted_fps.getvalue()
    unsorted_fps.close()
    unsorted_fps = None


    if not reorder or not metadata.num_bits:
        start_padding, end_padding, unsorted_arena = _chemfp.make_unsorted_aligned_arena(
            unsorted_arena, alignment)
        return FingerprintArena(metadata, alignment, start_padding, end_padding, storage_size,
                                unsorted_arena, "", ids,
                                num_bits=num_bits, num_bytes=num_bytes)

    # Reorder
        
    ordering = (ChemFPOrderedPopcount*len(ids))()
    popcounts = array.array("i", (0,)*(metadata.num_bits+2))

    start_padding, end_padding, unsorted_arena = _chemfp.make_sorted_aligned_arena(
        num_bits, storage_size, unsorted_arena, len(ids),
        ordering, popcounts, alignment)

    new_ids = [ids[item.index] for item in ordering]
    return FingerprintArena(metadata, alignment,
                            start_padding, end_padding, storage_size,
                            unsorted_arena, popcounts.tostring(), new_ids,
                            num_bits=num_bits, num_bytes=num_bytes)

# Coroutine to make an arena

def _make_arena_writer(metadata, reorder=True, alignment=8, num_bits=None, num_bytes=None):
    if num_bits is None:
        num_bits = metadata.num_bits
        
    if num_bytes is None:
        num_bytes = metadata.num_bytes
        if num_bytes is None:
            raise ValueError("Missing num_bytes")

    if num_bits is None:
        num_bits = num_bytes * 8
        
    if alignment is None:
        alignment = chemfp.arena.get_optimal_alignment(num_bits)

    storage_size = num_bytes
    if storage_size % alignment != 0:
        n = alignment - (storage_size % alignment)
        end_padding = b"\0" * n
        storage_size += n
    else:
        end_padding = b""

    ids = []
    unsorted_fps = BytesIO()


    while 1:
        id_fp_pairs = (yield "next")
        if id_fp_pairs is None:
            # Polite request to end.
            break

        for id, fp in id_fp_pairs:
            if len(fp) != num_bytes:
                raise ValueError("Fingerprint for id %r is %d bytes long, expected %d bytes: %r"
                                 % (ids[i], len(fp), num_bytes, fp))

            unsorted_fps.write(fp + end_padding)
            ids.append(id)

    # Convert to an arena
            
    unsorted_arena = unsorted_fps.getvalue()
    unsorted_fps.close()   # Reduce memory use
    unsorted_fps = None

    if not reorder:
        start_padding, end_padding, unsorted_arena = _chemfp.make_unsorted_aligned_arena(
            unsorted_arena, alignment)
        yield chemfp.arena.FingerprintArena(metadata, alignment, start_padding, end_padding,
                                            storage_size, unsorted_arena, b"", ids)
        return

    # Reorder

    ordering = (ChemFPOrderedPopcount*len(ids))()
    popcounts = array.array("i", (0,)*(num_bits+2))

    start_padding, end_padding, unsorted_arena = _chemfp.make_sorted_aligned_arena(
        num_bits, storage_size, unsorted_arena, len(ids),
        ordering, popcounts, alignment)

    new_ids = [ids[item.index] for item in ordering]
    yield FingerprintArena(metadata, alignment,
                           start_padding, end_padding, storage_size,
                           unsorted_arena, tobytes(popcounts), new_ids)

# Definitely not part of the public API.
# I think there needs to be a top-level function to create this.
# Perhaps 'chemfp.make_arena_builder()'?
# Or perhaps call it a MemoryWriter?
# I have 'add_fingerprint*()' here. Should it be 'write_fingerprint*()'?
class ArenaBuilder(object):
    def __init__(self, metadata, reorder=True, alignment=None, num_bits=None, num_bytes=None):
        self.metadata = metadata

        if num_bits is None:
            num_bits = metadata.num_bits
        if num_bytes is None:
            num_bytes = metadata.num_bytes

        if num_bits is None:
            if num_bytes is None:
                raise ValueError("Must specify at least one of num_bits or num_bytes, or define one in the metadata")
            num_bits = num_bytes * 8

        self._num_bits = num_bits

        self.reorder = reorder
        
        if alignment is None:
            alignment = get_optimal_alignment(num_bits)
        self.alignment = alignment

        self._writer = _make_arena_writer(metadata, reorder, alignment, num_bits)
        next(self._writer)  # prime the pump

    def add_fingerprint(self, id, fp):
        writer = self._writer
        if writer is None:
            raise ValueError("Cannot add a record after calling make_arena()")
        writer.send( [(id, fp)] )

    def add_fingerprints(self, id_fp_pairs):
        # Pass in an (id, fp) iterator
        writer = self._writer
        if writer is None:
            raise ValueError("Cannot add fingerprints after calling make_arena()")
        writer.send(id_fp_pairs)

    def make_arena(self):
        arena = next(self._writer)
        for x in self._writer:
            raise AssertionError
        self._writer = None
        return arena
        



# Copyright (c) 2010-2017 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import, print_function

# Information about fingerprint types
from . import argparse
from . import FingerprintIterator, Metadata

from . import io
from .encodings import import_decoder  # XXX too specific to the decoder module


def check_openbabel_maccs166():
    from .openbabel import HAS_MACCS, MACCS_VERSION
    assert HAS_MACCS
    if MACCS_VERSION == 1:
        return "OpenBabel-MACCS/1"
    elif MACCS_VERSION == 2:
        return "OpenBabel-MACCS/2"
    raise AssertionError

def check_openeye_maccs166():
    from .openeye import _maccs_version
    return "OpenEye-MACCS166/" + _maccs_version

def check_openeye_path():
    return "OpenEye-Path/2"

def check_rdkit_maccs():
    from .rdkit import MACCS_VERSION
    return "RDKit-MACCS166/"+MACCS_VERSION

def check_rdkit_fingerprint():
    from .rdkit import RDK_VERSION
    return "RDKit-Fingerprint/"+RDK_VERSION

def check_rdkit_atom_pair():
    from .rdkit import ATOM_PAIR_VERSION
    if ATOM_PAIR_VERSION is None:
        ATOM_PAIR_VERSION = "2"  # Nothing is supported; pretend to want v2
    return "RDKit-AtomPair/"+ATOM_PAIR_VERSION

def check_rdkit_torsion():
    from .rdkit import TORSION_VERSION
    return "RDKit-Torsion/"+TORSION_VERSION

def check_rdkit_pattern():
    from .rdkit import PATTERN_VERSION
    return "RDKit-Pattern/"+PATTERN_VERSION

### The chemfp fingerprint type API isn't powerful enough

# I have to list all of the possible fingerprint types, even if the
# platform doesn't support that specific type. There's also no support
# for toolkit vendors which do carefully tracks the fingerprint
# version (like OEChem); I should be using their version information
# rather than doing it myself.

_family_config_paths = (
    ("OpenEye-MACCS166/2", "chemfp.openeye.OpenEyeMACCSFingerprintFamily_v2"),
    ("OpenEye-MACCS166/3", "chemfp.openeye.OpenEyeMACCSFingerprintFamily_v3"),
    ("OpenEye-Path/2", "chemfp.openeye.OpenEyePathFingerprintFamily_v2"),
    ("OpenEye-Circular/2", "chemfp.openeye.OpenEyeCircularFingerprintFamily_v2"),
    ("OpenEye-Tree/2", "chemfp.openeye.OpenEyeTreeFingerprintFamily_v2"),
    
    ("RDKit-MACCS166/1", "chemfp.rdkit.RDKitMACCSFingerprintFamily_v1"),
    ("RDKit-MACCS166/2", "chemfp.rdkit.RDKitMACCSFingerprintFamily_v2"),
    ("RDKit-Fingerprint/1", "chemfp.rdkit.RDKitFingerprintFamily_v1"),
    ("RDKit-Fingerprint/2", "chemfp.rdkit.RDKitFingerprintFamily_v2"),
    ("RDKit-Morgan/1", "chemfp.rdkit.RDKitMorganFingerprintFamily_v1"),
    ("RDKit-Torsion/1", "chemfp.rdkit.RDKitTorsionFingerprintFamily_v1"),
    ("RDKit-Torsion/2", "chemfp.rdkit.RDKitTorsionFingerprintFamily_v2"),
    ("RDKit-AtomPair/1", "chemfp.rdkit.RDKitAtomPairFingerprintFamily_v1"),
    ("RDKit-AtomPair/2", "chemfp.rdkit.RDKitAtomPairFingerprintFamily_v2"),
    
    ("RDKit-Pattern/1", "chemfp.rdkit.RDKitPatternFingerprint_v1"),
    ("RDKit-Pattern/2", "chemfp.rdkit.RDKitPatternFingerprint_v2"),
    ("RDKit-Pattern/3", "chemfp.rdkit.RDKitPatternFingerprint_v3"),
    ("RDKit-Pattern/4", "chemfp.rdkit.RDKitPatternFingerprint_v4"),
    
    ("RDKit-Avalon/1", "chemfp.rdkit.RDKitAvalonFingerprint_v1"),
    
    ("OpenBabel-FP2/1", "chemfp.openbabel.OpenBabelFP2FingerprintFamily_v1"),
    ("OpenBabel-FP3/1", "chemfp.openbabel.OpenBabelFP3FingerprintFamily_v1"),
    ("OpenBabel-FP4/1", "chemfp.openbabel.OpenBabelFP4FingerprintFamily_v1"),
    ("OpenBabel-MACCS/1", "chemfp.openbabel.OpenBabelMACCSFingerprintFamily_v1"),
    ("OpenBabel-MACCS/2", "chemfp.openbabel.OpenBabelMACCSFingerprintFamily_v2"),

    ("Indigo-Similarity/1", "chemfp.indigo.IndigoSimilarityFingerprinter_v1"),
    ("Indigo-Substructure/1",
                      "chemfp.indigo.IndigoSubstructureFingerprinter_v1"),
    ("Indigo-ResonanceSubstructure/1",
                      "chemfp.indigo.IndigoResonanceSubstructureFingerprinter_v1"),
    ("Indigo-TautomerSubstructure/1",
                      "chemfp.indigo.IndigoTautomerSubstructureFingerprinter_v1"),
    ("Indigo-Full/1", "chemfp.indigo.IndigoFullFingerprinter_v1"),

    # In the future this will likely change to use a parameterized class
    # which can dynamically load fingerprint definitions

    ("ChemFP-Substruct-OpenEye/1",
                      "chemfp.openeye_patterns.SubstructOpenEyeFingerprinter_v1"),
    ("RDMACCS-OpenEye/1",
                      "chemfp.openeye_patterns.RDMACCSOpenEyeFingerprinter_v1"),
    ("RDMACCS-OpenEye/2",
                      "chemfp.openeye_patterns.RDMACCSOpenEyeFingerprinter_v2"),

    ("ChemFP-Substruct-RDKit/1",
                      "chemfp.rdkit_patterns.SubstructRDKitFingerprinter_v1"),
    ("RDMACCS-RDKit/1",
                      "chemfp.rdkit_patterns.RDMACCSRDKitFingerprinter_v1"),
    ("RDMACCS-RDKit/2",
                      "chemfp.rdkit_patterns.RDMACCSRDKitFingerprinter_v2"),

    ("ChemFP-Substruct-OpenBabel/1",
                      "chemfp.openbabel_patterns.SubstructOpenBabelFingerprinter_v1"),
    ("RDMACCS-OpenBabel/1",
                      "chemfp.openbabel_patterns.RDMACCSOpenBabelFingerprinter_v1"),
    ("RDMACCS-OpenBabel/2",
                      "chemfp.openbabel_patterns.RDMACCSOpenBabelFingerprinter_v2"),

    ("ChemFP-Substruct-Indigo/1",
                      "chemfp.indigo_patterns.SubstructIndigoFingerprinter_v1"),
    ("RDMACCS-Indigo/1",
                      "chemfp.indigo_patterns.RDMACCSIndigoFingerprinter_v1"),
)

_alternates = {
    "OpenEye-Path": check_openeye_path,
    "OpenBabel-MACCS": check_openbabel_maccs166,
    "OpenEye-MACCS166": check_openeye_maccs166,
    "RDKit-MACCS166": check_rdkit_maccs,
    "RDKit-Fingerprint": check_rdkit_fingerprint,
    "RDKit-AtomPair": check_rdkit_atom_pair,
    "RDKit-Torsion": check_rdkit_torsion,
    "RDKit-Pattern": check_rdkit_pattern,
    }


def _initialize_families(config_paths):
    d = {}
    for name, path in config_paths:
        # Set both the versioned and non-versioned names
        # The paths must be in order from oldest to newest
        
        unversioned_name = name.split("/")[0]
        d[unversioned_name] = d[name] = path
    return d

# Convert into a dictionary, and include the unversioned named
_family_config_paths = _initialize_families(_family_config_paths)

_loaded_families = {}

# Return the fingerprint family given a name like
# "OpenBabel-FP2" or "RDKit-Morgan/1"

def get_fingerprint_family(name):
    try:
        return _loaded_families[name]
    except KeyError:
        pass
    
    # Let's see if we can load it.

    # Is there a better name for this?
    try:
        func = _alternates[name]
    except KeyError:
        new_name = name
    else:
        new_name = func()

    try:
        path = _family_config_paths[new_name]
    except KeyError:
        raise ValueError("Unknown fingerprint family %r" % (name,))
    
    config = import_decoder(path)
    config.validate()
    
    family = FingerprintFamily(config)
    _loaded_families[name] = _loaded_families[new_name] = family
    return family

class FingerprintFamily(object):
    def __init__(self, config):
        self.config = config
        #name = config.name
        #format_string = config.format_string

    def __call__(self, **kwargs):
        return Fingerprinter(self.config, kwargs)

    def make_fingerprinter_from_type(self, type):
        terms = type.split()
        if not terms:
            raise ValueError("Empty fingerprint type (%r)" % (type,))

        required_args = self.config.get_args()

        kwargs = {}
        for term in terms[1:]:
            try:
                left, right = term.split("=")
            except ValueError:
                raise ValueError("Term %r in type %r must have one and only one '='" %
                                 (term, type))
            if left in kwargs:
                raise ValueError("Duplicate name %r in type %r" % (left, type))

            try:
                decoder = required_args[left].decoder
            except KeyError:
                raise ValueError("Unknown name %r in type %r" % (left, type))
            
            try:
                value = decoder(right)
            except ValueError, err:
                raise ValueError("Unable to parse %s value %r in type %r" % (
                    left, right, type))
            
            kwargs[left] = value

        # Fill in any missing default
        for name, arg in required_args.items():
            if name not in kwargs:
                kwargs[name] = arg.default

        # Let the configuration verify the kwargs
        verify_args = self.config.verify_args
        if verify_args is not None:
            verify_args(kwargs)

        return Fingerprinter(self.config, kwargs)


class Fingerprinter(object):
    def __init__(self, config, fingerprinter_kwargs):
        self.config = config
        if isinstance(config.num_bits, int):
            self.num_bits = config.num_bits
        elif config.num_bits is None:
            raise AssertionError("num_bits not defined for %r" % (config.name,))
        else:
            self.num_bits = config.num_bits(fingerprinter_kwargs)
        
        self.fingerprinter_kwargs = fingerprinter_kwargs

    def __eq__(self, other):
        return self.get_type() == other.get_type()

    def __ne__(self, other):
        return self.get_type() != other.get_type()

    def get_type(self):
        if self.config.format_string is None:
            assert not self.fingerprinter_kwargs, "kwargs but no format string!"
            return self.config.name
        encoded = self.config.format_string % self._encode_parameters()
        return self.config.name + " " + encoded

    def _encode_parameters(self):
        d = {}
        for k, v in self.fingerprinter_kwargs.items():
            encoder = self.config.args[k].encoder
            if encoder is None:
                d[k] = v
            else:
                d[k] = encoder(v)
        return d

    def read_molecule_fingerprints(self, source, format=None, id_tag=None, reader_args=None,
                                   errors="strict", metadata=None, location=None):
        source_filename = io.get_filename(source)
        if source_filename is None:
            sources = []
        else:
            sources = [source_filename]
            
        if metadata is None:
            # XXX I don't like how the user who wants to pass in aromaticity
            # information needs to create the full Metadata
            metadata = Metadata(num_bits=self.num_bits, type=self.get_type(),
                                software=self.config.software,
                                sources=sources)
            
        structure_reader = self.config.read_structures(metadata, source, format, id_tag,
                                                       reader_args, errors, location=location)
        fingerprinter = self.config.make_fingerprinter(**self.fingerprinter_kwargs)

        def fingerprint_reader(structure_reader, fingerprinter):
            for (id, mol) in structure_reader:
                yield id, fingerprinter(mol)
        reader = fingerprint_reader(structure_reader, fingerprinter)
        
        return FingerprintIterator(Metadata(num_bits = self.num_bits,
                                            sources = sources,
                                            software = self.config.software,
                                            type = self.get_type(),
                                            date = io.utcnow(),
                                            aromaticity = metadata.aromaticity),
                                   reader)

#    def describe(self, bitno):
#        if not (0 <= bitno < self.num_bits):
#            raise KeyError("Bit number %d is out of range" % (bitno,))
#
#        bit_description = self.config.bit_description
#        if bit_description is None:
#            return "(unknown)"
#        return bit_description[bitno]


## Some code to figure out what the format strings are inside of a string

class Dummy(object):
    def __str__(self):
        return ""
    def __int__(self):
        return 0

class GetArgs(object):
    def __init__(self, format_string):
        self.format_string = format_string
        self.args = []
    def __getitem__(self, name):
        # O(n**2) but n is small.
        if name in self.args:
            raise TypeError("Duplicate name %r in format string %r" % (name, self.format_string))
        self.args.append(name)
        return Dummy()

def _get_arg_names(s):
    if not s:
        return []
    args = GetArgs(s)
    s % args
    return args.args
    
class FingerprintArgument(object):
    def __init__(self, name, decoder, encoder, kwargs):
        self.name = name
        self.decoder = decoder
        self.encoder = encoder
        self.default = kwargs["default"]
        self.kwargs = kwargs

# Can't use "first or second" because some of the 'first' arguments
# can be false value like {}.
def OR(first, second):
    if first is None:
        return second
    return first
    
class FingerprintFamilyConfig(object):
    def __init__(self,
                 name = None,
                 format_string = None,
                 software = None,
                 num_bits = None,
                 read_structures = None,
                 make_fingerprinter = None,
                 verify_args = None,
                 args = None,
                 ):
        self.name = name
        self.format_string = format_string
        self.software = software
        self.num_bits = num_bits
        self.read_structures = read_structures
        self.make_fingerprinter = make_fingerprinter
        if args is None:
            args = {}
        self.verify_args = verify_args
        self.args = args.copy() # This can contain extra args!
        self._exact_args = None

    def validate(self):
        pass

    def get_args(self):
        args = self._exact_args
        if args is None:
            result = {}
            for name in _get_arg_names(self.format_string):
                result[name] = self.args[name]
            args = self._exact_args = result
        return args

    def clone(self, name=None, format_string=None, software=None,
              num_bits=None, read_structures=None, make_fingerprinter=None,
              verify_args=None, args=None):
        return FingerprintFamilyConfig(
            name = OR(name, self.name),
            format_string = OR(format_string, self.format_string),
            software = OR(software, self.software),
            num_bits = OR(num_bits, self.num_bits),
            read_structures = OR(read_structures, self.read_structures),
            make_fingerprinter = OR(make_fingerprinter, self.make_fingerprinter),
            verify_args = OR(verify_args, self.verify_args),
            args = OR(args, self.args))

    def add_argument(self, name, decoder=None, encoder=None, default=None,
                     action=None, metavar=None, help=None):
        #if name in self.args:
        #    raise AssertionError("Argument %r already added" % (name,))
        if default is not None:
            if help is not None:
                help = "%s (default=%s)" % (help, default)
        def parse_argument(s):
            try:
                return decoder(s)
            except ValueError, err:
                raise argparse.ArgumentError(None, "%s %s" % (name, err))
        arg = FingerprintArgument(name, decoder, encoder,
                                  kwargs = dict(type=parse_argument,
                                                default=default,
                                                action=action,
                                                metavar=metavar,
                                                help=help))
        self.args[name] = arg

    def remove_argument(self, name):
        del self.args[name]

    def add_argument_to_argparse(self, name, parser):
        info = self.args[name]
        parser.add_argument("--" + info.name, **info.kwargs)
        

# Helper functions

def positive_int(s):
    # Don't do int(s) because that allows "+3" and " 3 ", which I don't want
    if not s.isdigit():
        raise ValueError("must be 1 or greater")
    i = int(s)
    if i == 0:
        raise ValueError("must be 1 or greater")
    return i

def nonnegative_int(s):
    if not s.isdigit():
        raise ValueError("must be 0 or greater")
    return int(s)

def zero_or_one(s):
    if s == "0":
        return 0
    if s == "1":
        return 1
    raise ValueError("must be 0 or a 1")
        

def parse_type(type):
    terms = type.split()
    if not terms:
        raise ValueError("Empty fingerprint type (%r)" % (type,))

    name = terms[0]
    family = get_fingerprint_family(name)
    return family.make_fingerprinter_from_type(type)



# Copyright (c) 2010-2018 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

import argparse

import sys

import itertools
import heapq
import struct
import os
import re

from .. import __version__ as chemfp_version
from .. import arena as arena_module
from .. import bitops
from .. import io
from .. import open as _open_fingerprints, ChemFPError, Metadata, open_fingerprint_writer

from . import cmdsupport
from .._compat import myrepr

imap = getattr(itertools, "imap", map) # itertools.imap() for Python 2.7, map() for Python 3.x

def die(errmsg):
    sys.stderr.write("ERROR: %s\n" % (errmsg,))
    return SystemExit(1)

def open_fingerprints(source, format):
    try:
        return _open_fingerprints(source, format=format)
    except (IOError, io.ParseError) as err:
        # IOErrors and ParseErrors include the filename
        raise die(err)
    except (ChemFPError, ValueError) as err:
        assert source is not None, source  # Not supposed to raise this case
        raise die("%s: %r" % (err, source))

###### Progress classes

class _Progress(object):
    def __init__(self, outfile):
        self.outfile = outfile
        self._reset_msg = ""

    def start(self):
        pass
        
    def _update(self, msg):
        if msg is None:
            return
        self.outfile.write(self._reset_msg)
        self.outfile.write(msg)
        self.outfile.flush()
        self._reset_msg = "\r" + (" " *len(msg)) + "\r"
        
    def start_close(self):
        self._update(self.close_msg)
        
    def end(self):
        self.outfile.write(self._reset_msg)
        self.outfile.flush()
        self._reset_msg = ""

class SequentialProgress(_Progress):
    def __init__(self, num_filenames, close_msg=None, outfile=sys.stderr):
        _Progress.__init__(self, outfile)
        self.num_filenames = num_filenames
        self.close_msg = close_msg
        self.outfile = outfile
        self.count = 0
        if not num_filenames:
            self._where = " records from stdin."
        else:
            self._where = None  # You must first call 'start_filename'!

    def start_filename(self, filename):
        self.count += 1
        if self.num_filenames == 1:
            where = " records from %s" % (repr(filename)[1:-1])
        else:
            where = " records. File %d/%d: %s" % (self.count, self.num_filenames, repr(filename)[1:-1])
        self._where = where

    def update_status(self, num_records):
        msg = str(num_records) + self._where
        self._update(msg)


class MergeProgress(_Progress):
    def __init__(self, num_filenames, outfile=sys.stderr):
        assert num_filenames >= 2, "use a SequentialProgress"
        _Progress.__init__(self, outfile)
        self._where = " records from %d files." % (num_filenames,)

    def update_status(self, num_records):
        msg = str(num_records) + self._where
        self._update(msg)


###### Use when the output is in FPS format but where the fingerprints must be sorted
        
class MemoryWriter(object):
    def __init__(self, arena_builder, output, format):
        self.arena_builder = arena_builder
        self.output = output
        self.format = format

    def write_fingerprints(self, id_fp_pairs):
        self.arena_builder.add_fingerprints(id_fp_pairs)

    def close(self):
        arena = self.arena_builder.make_arena()
        self.arena_builder = None
        arena.save(self.output, format=self.format)
            

def make_memory_writer(output, metadata, format):
    # Want to reorder, and want fps output. Do this in memory
    arena_builder = arena_module.ArenaBuilder(metadata=metadata, reorder=True, alignment=1)
    return MemoryWriter(arena_builder, output, format)
                              

###### 
            

def make_sequential_chunk_reader(filenames, format, progress):
    progress.start_filename(filenames[0])
    compat_values, compat_filenames = start_compatibility()

    reader = open_fingerprints(filenames[0], format=format)
    try:
        check_compatibility(compat_values, compat_filenames, reader, filenames[0])
    except TypeError as err:
        raise die(str(err))

    def read_sequential_records(first_reader, filenames, format):
        yield first_reader
        first_reader.close()
        first_reader = None
        for i, filename in enumerate(filenames):
            progress.start_filename(filename)
            reader = open_fingerprints(filename, format=format)
            try:
                check_compatibility(compat_values, compat_filenames, reader, filename)
            except TypeError as err:
                reader.close()
                raise die(str(err))
            yield reader
            reader.close()

    # Use information from the first reader to make a new metadata.
    md = reader.metadata
    metadata = Metadata(
        num_bits = md.num_bits,
        num_bytes = md.num_bytes,
        type = md.type,
        aromaticity = md.aromaticity,
        software = md.software,
        #date = chemfp.io.utcnow(),
        )
        
         
    return metadata, read_sequential_records(reader, filenames[1:], format)

# Keep track of a dictionary of expected values for each compatability factor.
def start_compatibility():
    compat_values = {}
    compat_filenames = {}
    for name in ("type", "num_bits", "num_bytes", "storage_size", "software", "aromaticity"):
        compat_values[name] = None      # The expected value; None if unknown, _incompatible if problem.
        compat_filenames[name] = None   # Where the information came from
    compat_values["all_fpb"] = True
    return compat_values, compat_filenames


_incompatible = object()

# Check a given attribute for compatibility issues.
def _compatibility(attribute, values, filenames,
                   new_value, new_filename, error):
    # Compare to an old value, if it exists.
    old_value = values[attribute]
    if old_value is None:
        if new_value is None:
            # Nothing assigned
            return
        # Use the new value
        values[attribute] = new_value
        filenames[attribute] = new_filename
        return

    if new_value is None:
        # Have an old value, don't have a new value. Stay with the old value.
        return

    # Have an old and a new value. Make sure they are the same.
    if old_value != new_value:
        if old_value is _incompatible:  # already seen this. Supress additional warnings.
            return
        old_filename = filenames[attribute]
        values[attribute] = _incompatible
        error(values, attribute, old_value, old_filename, new_value, new_filename)
    
def _compat_error(values, attribute, old_value, old_filename, new_value, new_filename):
    raise TypeError("%r has a %s of %s, which is not compatible with the %r which has a %s of %s"
                    % (new_filename, attribute, myrepr(new_value), old_filename, attribute, myrepr(old_value)))

def _compat_ignore(values, attribute, old_value, old_filename, new_value, new_filename):
    pass

## def _compat_first(values, attribute, old_value, old_filename, new_value, new_filename):
##     # Restore the old value
##     values[attribute] = old_value

def _compat_warning(values, attribute, old_value, old_filename, new_value, new_filename):
    sys.stderr.write("WARNING: %r has a %s of %s, which is not compatible with %r which has a %s of %s"
                     % (new_filename, attribute, myrepr(new_value), old_filename, attribute, myrepr(old_value)))
    


def check_compatibility(compat_values, compat_filenames, reader, new_filename):
    metadata = reader.metadata
    _compatibility("type", compat_values, compat_filenames,
                   metadata.type, new_filename, _compat_error)
    #print >>open("/dev/stderr", "w"), "Check", metadata, repr(metadata), repr(metadata.num_bytes)
    if metadata.num_bytes is not None:
        # By construction, this means the file is empty, and without metadata
        _compatibility("num_bits", compat_values, compat_filenames,
                       metadata.num_bits, new_filename, _compat_error)
        _compatibility("num_bytes", compat_values, compat_filenames,
                       metadata.num_bytes, new_filename, _compat_error)
    _compatibility("software", compat_values, compat_filenames,
                   metadata.software, new_filename, _compat_ignore)
    _compatibility("aromaticity", compat_values, compat_filenames,
                   metadata.aromaticity, new_filename, _compat_warning)  # use logs?

    if isinstance(reader, arena_module.FingerprintArena):
        _compatibility("storage_size", compat_values, compat_filenames,
                       reader.storage_size, new_filename, _compat_ignore)
    else:
        compat_values["all_fpb"] = False

################ Merge multiple inputs in parallel

# Assume the inputs are in popcount order, and keep the output
# fingerprints in popcount order. (Break ties based on input filename
# index.)

# Standard decorate-sort-undecorate approach, using heapq.merge to
# sort the already-sorted inputs.
         
def decorate_with_popcount(reader, position):
    byte_popcount = bitops.byte_popcount
    for id, fp in reader:
        yield byte_popcount(fp), position, id, fp

def decorate_arena_with_popcount(arena, position):
    popcount_indices = arena.popcount_indices
    indices = struct.unpack("<" + "I"*(len(popcount_indices)//4), popcount_indices)
    for popcount in range(len(indices)-1):
        start_index = indices[popcount]
        end_index = indices[popcount+1]
        subarena = arena[start_index:end_index]
        for id, fp in subarena:
            yield popcount, position, id, fp


from operator import itemgetter
undecorate = itemgetter(2, 3)


def make_merge_chunk_reader(filenames, format):
    # Open multiple files in parallel and merge assuming sorted by popcount
    merge_info = Metadata()

    compat_values, compat_filenames = start_compatibility()

    readers = []
    decorated_readers = []
    sources = []
    for position, filename in enumerate(filenames):
        reader = open_fingerprints(filename, format=format)
        readers.append(reader)
        try:
            check_compatibility(compat_values, compat_filenames, reader, filename)
        except TypeError as err:
            # Close all currently open files
            for r in readers:
                r.close()
            raise die(str(err))
        sources.extend(reader.metadata.sources)         # merge all sources

        if isinstance(reader, arena_module.FingerprintArena) and reader.popcount_indices:
            # Easy optimization to decorate this case
            decorated_reader = decorate_arena_with_popcount(reader, position)
        else:
            # Need to do the byte_popcount for this case.
            decorated_reader = decorate_with_popcount(reader, position)
            
        decorated_readers.append(decorated_reader)

    software = compat_values["software"]
    if software is _incompatible:
        software = None

    aromaticity = compat_values["aromaticity"]
    if aromaticity is _incompatible:
        aromaticity = None

    metadata = Metadata(
        num_bits = compat_values["num_bits"],
        num_bytes = compat_values["num_bytes"],
        sources = sources,
        software = software,
        aromaticity = aromaticity,
        #date = datetime.datetime.now(),
        #date = io.utcnow(),
        )

        
    def merge_chunk_reader(readers, id_fp_pairs):
        try:
            yield id_fp_pairs
        finally:
            for reader in readers:
                reader.close()

    return metadata, merge_chunk_reader(readers, imap(undecorate, heapq.merge(*decorated_readers)))
    

epilog = """

Examples:

fpcat can be used to merge multiple FPS files. For example, you might
have used GNU parallel to generate FPS files for each of the PubChem
files, which you want to merge into a single file.:

    fpcat Compound_*.fps -o pubchem.fps

The --merge option is experimental. Use it if the input fingerprints
are in popcount order, because sorted output is a simple merge sort of
the individual sorted inputs. However, this option opens all input
files at the same time, which may exceed your resource limit on file
descriptors. The current implementation also requires a lot of disk
seeks so is slow for many files.

    
"""



parser = argparse.ArgumentParser(
    description="Combine multiple fingerprint files into a single file.",
    epilog=epilog,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    )
parser.add_argument(
    "--in", metavar="FORMAT", dest="input_format", choices=("fps", "fps.gz", "flush"),
    help="input fingerprint format. One of fps or fps.gz. (default guesses from filename or is fps)")
parser.add_argument(
    "--merge", action="store_true",
    help="assume the input fingerprint files are in popcount order and do a merge sort")
parser.add_argument(
    "-o", "--output", metavar="FILENAME",
    help="save the fingerprints to FILENAME (default=stdout)")

parser.add_argument(
    "--out", metavar="FORMAT", dest="output_format", choices=("fps", "fps.gz", "flush"),
    help="output fingerprint format. One of fps or fps.gz. (default guesses from output filename, or is 'fps')")
parser.add_argument(
    "--reorder", action="store_true",
    help="reorder the output fingerprints by popcount")
parser.add_argument(
    "--preserve-order", action="store_true",
    help="save the output fingerprints in the same order as the input (default for FPS output)")

parser.add_argument(
    "--show-progress", action="store_true",
    help="show progress")

_scales = {
    "K" : 1000,
    "KB": 1000,
    "M" : 1000*1000,
    "MB": 1000*1000,
    "G" : 1000*1000*1000,
    "GB": 1000*1000*1000,
    "T" : 1000*1000*1000*1000,
    "TB": 1000*1000*1000*1000,
    "P" : 1000*1000*1000*1000*1000,
    "PB": 1000*1000*1000*1000*1000,
    }

_leading_digit = re.compile(r"(\d+(\.\d*)?)(KB?|MB?|GB?|TB?|PB?)?$")

_MIN_SIZE = 20*1000*1000
_MIN_SIZE_STR = "20MB"

def fancy_size(size):
    s = size.upper()
    if s == "NONE":
        return None
    m = _leading_digit.match(s)
    if not m:
        raise argparse.ArgumentTypeError("Unable to convert %r into a memory size" % (size,))

    scalar = m.group(1)
    scale = m.group(3)
    if not scale:
        scale = 1
    else:
        scale = _scales[scale]
        
    if "." in scalar:
        # Could goo back/forth through float, but since I
        # have the decimal module, I'll use it.
        import decimal
        result = int(decimal.Decimal(scalar) * scale)
    else:
        result = int(scalar) * scale

    if result < _MIN_SIZE:
        raise argparse.ArgumentTypeError("Must be at least %s, not %s bytes" % (_MIN_SIZE_STR, size))
    return result
    
cmdsupport.add_version(parser)
parser.add_argument("filenames", metavar="filename", nargs="*",
                    help="input fingerprint filenames (default: use stdin)")


def run():
    cmdsupport.run(main)

def main(args=None):
    args = parser.parse_args(args)

    if args.preserve_order and args.reorder:
        parser.error("Cannot specify both --preserve-order and --reorder; they are incompatible")

    if not cmdsupport.has_chemfp_converters:
        if args.input_format == "flush":
            cmdsupport.die("--in format 'flush' not supported because the chemfp_converter module is not available")
        if args.output_format == "flush":
            cmdsupport.die("--out format 'flush' not supported because the chemfp_converter module is not available")
        
    # Do some checking before I start doing I/O
    if not args.filenames:
        # Read from stdin, which means that only FPS is supported.
        if args.input_format == "fpb":
            raise die("The FPB input reader only works on memory-mapped files, and not stdin")
    else:
        # This is a preliminary check, used to give early feedback.
        for filename in args.filenames:
            if not os.path.exists(filename):
                raise die("Fingerprint file %r does not exist" % (filename,))


    output = args.output
    output_format = args.output_format
    if output is None:
        # Save to stdout, which means that only FPS is supported.
        if output_format == "fpb":
            raise die("The FPB output writer only works with seekable files, and not stdout")
    else:
        # Note: if the filename is "abc.fpb", then the dirname is ""
        # Normally I prefer to just open and catch the error. But with a
        # memory writer, the open doesn't occur until the very end, so
        # this error may be too late.
        dirname = os.path.dirname(output)
        if dirname:
            if not os.path.exists(dirname):
                raise die("The output directory for %r does not exist" % (dirname,))
            if not os.path.isdir(dirname):
                raise die("The output directory for %r is not actually a directory" % (dirname,))
        
        format, compression = io.normalize_output_format(output, output_format, ("fps", ""))
        if format == "fpb":
            raise die("This version of chemfp does not support the FPB format.")
        if compression:
            if format == "fps":
                output_format = format+"."+compression
            elif format == "flush":
                raise die("Compression of flush files is not supported")
            else:
                raise die("Unsupported output fingerprint format %r" % (format,))
        else:
            output_format = format
            if output_format not in ("fpb", "fps", "fps.gz", "flush"):
                raise die("Unsupported output fingerprint format %r" % (format,))
        

    if args.reorder:
        reorder = True
    elif args.preserve_order:
        reorder = False
    else:
        # Select a default based on the output type
        if output_format == "fpb":
            reorder = True
        else:
            reorder = False

    # TODO: Check if the inputs and the output are in fpb format.  If
    # so, and if the inputs all have the same num_bytes and
    # storage_size, then there are faster ways to merge.

    # Create the input reader

    class progress:        # Placeholder to support an output-specific 'close' message
        close_msg = None   # so I don't have to worry if an input-specific progress is available.

    # NOTE: the FPS reader reads the metadata before returning
    if not args.filenames:
        reader = open_fingerprints(None, format=args.input_format)
        progress = SequentialProgress(len(args.filenames), outfile=sys.stderr)
        def stdin_chunk_reader(reader):
            yield reader
        metadata, chunk_reader = reader.metadata, stdin_chunk_reader(reader)
    else:
        if args.merge and len(args.filenames) >= 2:
            progress = MergeProgress(len(args.filenames), outfile=sys.stderr)
            metadata, chunk_reader = make_merge_chunk_reader(args.filenames, format=args.input_format)
        else:
            progress = SequentialProgress(len(args.filenames), outfile=sys.stderr)
            metadata, chunk_reader = make_sequential_chunk_reader(args.filenames, format=args.input_format,
                                                                  progress=progress)

    # Create the output writer
            
    try:
        # Always try to open the writer now so I can catch any errors early.
        try:
            writer = open_fingerprint_writer(output, metadata=metadata, format=output_format)
        except ValueError as err:
            raise die("Cannot open fingerprint writer: %s" % (err,))
        progress.close_msg = ""
        
        if reorder:
            # Close now, and use save() in the memory writer to actually save it.
            # I could pass in the already-open writer, but that would bypass the
            # fast code to save an arena as FPB files for chemfp 2.x or later.
            # I want even chemfp 1.x to have the same double-open behavior.
            writer.close()
            # Build an arena into memory then save the arena
            writer = make_memory_writer(output, metadata=metadata, format=output_format)
            progress.close_msg = "Writing fingerprints in FPS format, ordered by popcount."

    except IOError as err:
        chunk_reader.close()
        raise die("Unable to open fingerprint writer: %s" % (err,))


    try:
        if args.show_progress:
            # Progress is hard because I want to show some progress even
            # when the input is from a single file or when merging files.
            # (In both cases I don't know the total number of records.)

            progress.start()
            num_processed = 0
            progress.update_status(num_processed)
            for id_fp_pairs in chunk_reader:
                id_fp_pairs = iter(id_fp_pairs) # make sure it's an iterator, so I can ...
                while 1:
                    id_fp_pairs_list = list(itertools.islice(id_fp_pairs, 50000)) # .. read 50,000 at a time
                    if not id_fp_pairs_list:
                        break
                    writer.write_fingerprints(id_fp_pairs_list)
                    num_processed += len(id_fp_pairs_list)
                    progress.update_status(num_processed)
            progress.start_close()
            writer.close()
            progress.end()
            if args.filenames:
                if len(args.filenames) == 1:
                    where =  "1 file"
                else:
                    where = "%d files" % (len(args.filenames),)
            else:
                where = "stdin"
            sys.stderr.write("Done. Processed %d records from %s.\n" % (num_processed, where))
            sys.stderr.flush()

        else:
            for id_fp_pairs in chunk_reader:
                writer.write_fingerprints(id_fp_pairs)
            writer.close()

    except (ChemFPError, IOError) as err:
        chunk_reader.close()
        raise die("Unable to create fingerprints:" + str(err))

    

if __name__ == "__main__":
    main(sys.argv[1:])



# This tests check for memory leaks.

import unittest2

import resource
import sys
import array

import _chemfp
import chemfp

def get_memory():
    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss

def memory_growth(f, *args, **kwargs):
    x = f(*args, **kwargs)
    m1 = get_memory()
    count = 0
    
    for i in xrange(1000):
        x = f(*args, **kwargs)
        m2 = get_memory()
        if m2 > m1:
            count += 1
            m1 = m2
    #print "count", count
    if count > 900:
        raise AssertionError("Memory growth in %r (%d/%d)!" % (f, count, 1000))
    return x

def fps_parse_id_fp():
    memory_growth(_chemfp.fps_parse_id_fp, 4, "ABCD\tSpam\n")


# I mistakenly used BuildValue("O") instead of BuildValue("N").
# This added an extra refcount to some of the returned strings,
# which meant they would never become freed.
class TestBuildValueMemoryLeaks(unittest2.TestCase):
    def test_make_unsorted_aligned_arena_when_aligned(self):
        s = "SpamAndEggs!"
        n1 = sys.getrefcount(s)
        memory_growth(_chemfp.make_unsorted_aligned_arena, s, 4)
        n2 = sys.getrefcount(s)
        if n2 > n1 + 10:
            raise AssertionError("Aligned growth has an extra refcount")

    def test_make_unsorted_aligned_arena_when_unaligned(self):
        SIZE = 1024*8
        s = "S" * SIZE
        n1 = sys.getrefcount(s)
        memory_growth(_chemfp.make_unsorted_aligned_arena, s, SIZE)
        n2 = sys.getrefcount(s)
        if n2 > n1 + 10:
            raise AssertionError("Aligned growth has an extra refcount: %d, %d" % (n1, n2))

    def test_align_fingerprint_same_size(self):
        s = "abcd"
        n1 = sys.getrefcount(s)
        memory_growth(_chemfp.align_fingerprint, s, 4, 4)
        n2 = sys.getrefcount(s)
        if n2 > n1 + 10:
            raise AssertionError("align_fingerprint with alignment has extra refcount: %d, %d" %
                                 (n1, n2))

    def test_align_fingerprint_needs_new_string(self):
        x = memory_growth(_chemfp.align_fingerprint, "ab", 16, 1024)
        # I don't know why the above doesn't find the leak.
        # I can probe the count directly.
        i = sys.getrefcount(x[2])
        if i != 2:
            raise AssertionError("Unexpected refcount: %d" % (i,))
    
    def test_make_sorted_aligned_arena_trivial(self):
        ordering = array.array("c", "\0\0\0\0"*16)  # space for a ChemFPOrderedPopcount (and extra)
        popcounts = array.array("c", "\0\0\0\0"*36)  # num_fingerprints + 2 (and extra)
        arena = "BLAH"
        n1 = sys.getrefcount(arena)
        memory_growth(_chemfp.make_sorted_aligned_arena, 32, 4, arena, 0, ordering, popcounts, 4)
        n2 = sys.getrefcount(arena)
        if n1 != n2:
            # This one doesn't need "N" because it borrowed the input refcount.
            raise AssertionError("Borrowed refcount should have been okay.")

    def test_make_sorted_aligned_arena_already_sorted_and_aligned(self):
        ordering = array.array("c", "\0\0\0\0"*16)  # space for a ChemFPOrderedPopcount (and extra)
        popcounts = array.array("c", "\0\0\0\0"*36)  # num_fingerprints + 2 (and extra)
        arena = "BDFL"
        n1 = sys.getrefcount(arena)
        memory_growth(_chemfp.make_sorted_aligned_arena, 32, 4, arena, 1, ordering, popcounts, 4)
        n2 = sys.getrefcount(arena)
        if n1 != n2:
            # This one doesn't need "N" because it borrowed the input refcount.
            raise AssertionError("Borrowed refcount should have been okay. %d != %d" % (n1, n2))

    def test_make_sorted_aligned_arena_already_sorted_but_not_aligned(self):
        ordering = array.array("c", "\0\0\0\0"*16)  # space for a ChemFPOrderedPopcount (and extra)
        popcounts = array.array("c", "\0\0\0\0"*40)  # num_fingerprints + 2 (and extra)
        arena = "QWOP" + "\0" * (4096-4)
        n1 = sys.getrefcount(arena)
        memory_growth(_chemfp.make_sorted_aligned_arena, 32, 4096, arena, 1, ordering, popcounts, 1024)
        n2 = sys.getrefcount(arena)
        if n1 != n2:
            # This one shouldn't touch the input arena
            raise AssertionError("Borrowed refcount should have been okay. %d != %d" % (n1, n2))

    def test_make_sorted_aligned_arena_when_not_sorted(self):
        ordering = array.array("c", "\0\0\0\0"*16)  # space for a ChemFPOrderedPopcount (and extra)
        popcounts = array.array("c", "\0\0\0\0"*40)  # num_fingerprints + 2 (and extra)
        arena = "QWON" + "\0" * (4096-4)
        n1 = sys.getrefcount(arena)
        memory_growth(_chemfp.make_sorted_aligned_arena, 32, 4, arena, 2, ordering, popcounts, 1024)
        n2 = sys.getrefcount(arena)
        if n1 != n2:
            # This one shouldn't touch the input arena
            raise AssertionError("Borrowed refcount should have been okay. %d != %d" % (n1, n2))

        
    def test_arena_copy(self):
        data = "ABCD\t0\n" * 200
        from cStringIO import StringIO
        arena = chemfp.load_fingerprints(StringIO(data))
        def make_subarena_copy():
            arena[1:].copy()
        memory_growth(make_subarena_copy)
    
if __name__ == "__main__":
    unittest2.main()



# Tests for the different alignment methods in bitops

from __future__ import absolute_import, with_statement
import unittest2
from cStringIO import StringIO

from support import fullpath

import chemfp
import chemfp.bitops
from chemfp.bitops import hex_decode
import _chemfp

set_alignment_method = chemfp.bitops.set_alignment_method
get_alignment_method = chemfp.bitops.get_alignment_method

CHEBI_TARGETS = fullpath("chebi_rdmaccs.fps")
CHEBI_QUERIES = fullpath("chebi_queries.fps.gz")

targets = chemfp.load_fingerprints(CHEBI_TARGETS, alignment=8)
targets_64 = chemfp.load_fingerprints(CHEBI_TARGETS, alignment=64)

available_methods = chemfp.bitops.get_methods()
alignment_methods = chemfp.bitops.get_alignment_methods()


all_methods = dict.fromkeys("LUT8-1 LUT8-4 LUT16-4 Lauradoux POPCNT Gillies ssse3".split())

class TestMethods(unittest2.TestCase):
    def test_no_duplicates(self):
        methods = chemfp.bitops.get_methods()
        self.assertEquals(len(methods), len(set(methods)))

    def test_for_unknown_methods(self):
        for method in chemfp.bitops.get_methods():
            self.assertIn(method, all_methods)

    def test_for_possible_missing_popcnt(self):
        if len(all_methods) == 4:
            self.assertNotIn("POPCNT", chemfp.get_methods())
        
    def test_internal_bad_args(self):
        with self.assertRaisesRegexp(IndexError, "method index is out of range"):
            _chemfp.get_method_name(-1)
        with self.assertRaisesRegexp(IndexError, "method index is out of range"):
            _chemfp.get_method_name(_chemfp.get_num_methods())

all_alignments = dict.fromkeys("align1 align4 align8-small align8-large align-ssse3".split())

class TestAlignments(unittest2.TestCase):
    def test_no_duplicates(self):
        alignments = chemfp.bitops.get_alignments()
        self.assertEquals(len(alignments), len(set(alignments)))
        self.assertEquals(len(alignments), len(all_alignments))

    def test_for_unknown_alignments(self):
        for alignment in chemfp.bitops.get_alignments():
            self.assertIn(alignment, all_alignments)

    def test_get_set_alignment_method(self):
        for alignment in chemfp.bitops.get_alignments():
            method = get_alignment_method(alignment)
            self.assertIn(method, all_methods)
            set_alignment_method(alignment, "LUT8-1")
            self.assertEqual(get_alignment_method(alignment), "LUT8-1")
            set_alignment_method(alignment, method)
            self.assertEqual(get_alignment_method(alignment), method)
            
        
    def test_internal_bad_args(self):
        with self.assertRaisesRegexp(IndexError, "alignment index is out of range"):
            _chemfp.get_alignment_name(-1)
        with self.assertRaisesRegexp(IndexError, "alignment index is out of range"):
            _chemfp.get_alignment_name(_chemfp.get_num_methods())


        # I didn't want a better error code for this
        with self.assertRaisesRegexp(ValueError, "Bad argument"):
            _chemfp.get_alignment_name(_chemfp.get_alignment_method(-1))
        with self.assertRaisesRegexp(ValueError, "Bad argument"):
            _chemfp.get_alignment_name(_chemfp.get_alignment_method(100))

        with self.assertRaisesRegexp(ValueError, "Bad argument"):
            _chemfp.get_alignment_name(_chemfp.set_alignment_method(-1, 0))
        with self.assertRaisesRegexp(ValueError, "Bad argument"):
            _chemfp.get_alignment_name(_chemfp.set_alignment_method(100, 0))


    def test_cannot_use_64_bit_method_for_shorter_bit_alignment(self):
        msg = "Mismatch between popcount method and alignment type"
        available_methods = chemfp.bitops.get_methods()
        for method in ("Lauradoux", "Gillies", "POPCNT"):
            if (method == "POPCNT") and ("POPCNT" not in available_methods):
                continue
            with self.assertRaisesRegexp(ValueError, msg):
                set_alignment_method("align1", method)
            with self.assertRaisesRegexp(ValueError, msg):
                set_alignment_method("align4", method)


    @unittest2.skipIf("ssse3" not in available_methods, "CPU does not implement SSSE3")
    def test_ssse3(self):
        method = get_alignment_method("align-ssse3")
        # This disables SSSE3 support
        set_alignment_method("align-ssse3", "LUT8-1")
        self.assertEquals(get_alignment_method("align-ssse3"), "LUT8-1")
        set_alignment_method("align-ssse3", "ssse3")
        self.assertEquals(get_alignment_method("align-ssse3"), "ssse3")
        set_alignment_method("align-ssse3", method)



class TestAlign8SmallMethods(unittest2.TestCase):
    def setUp(self):
        self.small_method = get_alignment_method("align8-small")
        self.large_method = get_alignment_method("align8-large")
        self.ssse3_method = get_alignment_method("align-ssse3")
    def tearDown(self):
        set_alignment_method("align8-small", self.small_method)
        set_alignment_method("align8-large", self.large_method)
        set_alignment_method("align-ssse3", self.ssse3_method)
        
    def _doit(self, method):
        for alignment in ("align8-small", "align8-large", "align-ssse3"):
            set_alignment_method(alignment, method)
            self.assertEquals(get_alignment_method(alignment), method)
        
        hits = targets.knearest_tanimoto_search_fp(hex_decode("00000000100410200290000b03a29241846163ee1f"), k=12, threshold=0.2).get_ids_and_scores()
        self.assertEqual(hits, [('CHEBI:8069', 1.0),
                                ('CHEBI:6758', 0.78723404255319152),
                                ('CHEBI:7983', 0.73999999999999999),
                                ('CHEBI:8107', 0.6956521739130435),
                                ('CHEBI:17568', 0.6904761904761905),
                                ('CHEBI:16294', 0.6818181818181818),
                                ('CHEBI:16964', 0.673469387755102),
                                ('CHEBI:17477', 0.6458333333333334),
                                ('CHEBI:17025', 0.62),
                                ('CHEBI:15901', 0.6122448979591837),
                                ('CHEBI:16742', 0.6122448979591837),
                                ('CHEBI:4888', 0.6078431372549019)])

    def test_lut8_1(self):
        self._doit("LUT8-1")
        
    def test_lut8_4(self):
        self._doit("LUT8-4")
        
    def test_lut16_4(self):
        self._doit("LUT16-4")

    def test_lauradoux(self):
        with self.assertRaisesRegexp(ValueError, "Mismatch between popcount method and alignment type"):
            set_alignment_method("align8-small", "Lauradoux")

    @unittest2.skipIf("POPCNT" not in alignment_methods, "CPU does not implement POPCNT")
    def test_popcnt(self):
        self._doit("POPCNT")

class TestAlign8LargeMethods(unittest2.TestCase):
    def setUp(self):
        self.large_method = get_alignment_method("align8-large")
        self.ssse3_method = get_alignment_method("align-ssse3")
    def tearDown(self):
        set_alignment_method("align8-large", self.large_method)
        set_alignment_method("align-ssse3", self.ssse3_method)
        
    def _doit(self, method, use_ssse3=False):
        set_alignment_method("align8-large", method)
        self.assertEquals(get_alignment_method("align8-large"), method)
        if use_ssse3:
            set_alignment_method("align-ssse3", "ssse3")
            self.assertEquals(get_alignment_method("align-ssse3"), "ssse3")
        else:
            set_alignment_method("align-ssse3", "LUT8-1")
            self.assertEquals(get_alignment_method("align-ssse3"), "LUT8-1")
        
        hits = targets_64.knearest_tanimoto_search_fp(hex_decode("00000000100410200290000b03a29241846163ee1f"), k=12, threshold=0.2).get_ids_and_scores()
        self.assertEqual(hits, [('CHEBI:8069', 1.0),
                                ('CHEBI:6758', 0.78723404255319152),
                                ('CHEBI:7983', 0.73999999999999999),
                                ('CHEBI:8107', 0.6956521739130435),
                                ('CHEBI:17568', 0.6904761904761905),
                                ('CHEBI:16294', 0.6818181818181818),
                                ('CHEBI:16964', 0.673469387755102),
                                ('CHEBI:17477', 0.6458333333333334),
                                ('CHEBI:17025', 0.62),
                                ('CHEBI:15901', 0.6122448979591837),
                                ('CHEBI:16742', 0.6122448979591837),
                                ('CHEBI:4888', 0.6078431372549019)])

    def test_lut8_1(self):
        self._doit("LUT8-1")
        
    def test_lut8_4(self):
        self._doit("LUT8-4")
        
    def test_lut16_4(self):
        self._doit("LUT16-4")
        
    def test_lauradoux(self):
        self._doit("Lauradoux")
        
    def test_gillies(self):
        self._doit("Gillies")

    @unittest2.skipIf("ssse3" not in available_methods, "CPU does not implement SSSE3")
    def test_ssse3(self):
        self._doit("Lauradoux", use_ssse3=True)

    @unittest2.skipIf("POPCNT" not in available_methods, "CPU does not implement POPCNT")
    def test_popcnt(self):
        self._doit("POPCNT")
        

class TestSelectFastestMethod(unittest2.TestCase):
    def setUp(self):
        self._alignment_methods = chemfp.bitops.get_alignment_methods()
    def tearDown(self):
        for k,v in self._alignment_methods.items():
            set_alignment_method(k, v)
            
    def test_select_fastest(self):
        for alignment in all_alignments:
            set_alignment_method(alignment, "LUT8-1")
            self.assertEquals(get_alignment_method(alignment), "LUT8-1")

        chemfp.bitops.select_fastest_method()

        best_methods1 = chemfp.bitops.get_alignment_methods()

        for alignment in all_alignments:
            set_alignment_method(alignment, "LUT8-1")
            self.assertEquals(get_alignment_method(alignment), "LUT8-1")

        chemfp.bitops.select_fastest_method()

        best_methods2 = chemfp.bitops.get_alignment_methods()
        self.assertEquals(best_methods1, best_methods2) # This might fail if two methods have nearly identical timings

# Tests based on coverage analysis
class TestErrorConditions(unittest2.TestCase):
    def test_get_unknown_alignment(self):
        with self.assertRaisesRegexp(ValueError, "Unknown alignment 'Blah'"):
            get_alignment_method("Blah")
    def test_set_unknown_alignment(self):
        with self.assertRaisesRegexp(ValueError, "Unknown alignment 'Blah'"):
            set_alignment_method("Blah", "LUT8-1")
    def test_set_unknown_alignment_method(self):
        with self.assertRaisesRegexp(ValueError, "Unknown method 'LUT8-7'"):
            set_alignment_method("align1", "LUT8-7")

    def test_too_large_repeat(self):
        with self.assertRaisesRegexp(ValueError, "repeat size is meaninglessly large"):
            chemfp.bitops.select_fastest_method(2**30)
    def test_negative_repeat(self):
        with self.assertRaisesRegexp(ValueError, "repeat size must be 1 or larger.*"):
            chemfp.bitops.select_fastest_method(-10)

class TestPrintReport(unittest2.TestCase):
    def test_print_report(self):
        stdout = StringIO()
        chemfp.bitops.print_report(stdout)
        output = stdout.getvalue()

class TestEnvironmentVariables(unittest2.TestCase):
    def setUp(self):
        self._alignment_methods = chemfp.bitops.get_alignment_methods()
    def tearDown(self):
        for k,v in self._alignment_methods.items():
            set_alignment_method(k, v)

    def set_align4(self):
        set_alignment_method("align4", "LUT8-4")
        chemfp.bitops.use_environment_variables({"CHEMFP-ALIGN4": "LUT8-1"})
        self.assertEquals(get_alignment_methods("align4"), "LUT8-1")

if __name__ == "__main__":
    chemfp.bitops.use_environment_variables()
    unittest2.main()
    



"""Create OpenEye fingerprints


"""

# Copyright (c) 2010-2017 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import

import sys
import os
import errno
import ctypes
import warnings

from openeye.oechem import *
from openeye.oegraphsim import *

from . import ParseError
from . import types
from . import io
from . import SOFTWARE as CHEMFP_SOFTWARE

__all__ = ["read_structures", "get_path_fingerprinter", "get_maccs_fingerprinter"]


class UnknownFormat(KeyError):
    def __str__(self):
        return "Unknown format %r" % (self.args[0],)

############# Used when generate the FPS header

SOFTWARE = "OEGraphSim/%(release)s (%(version)s)" % dict(
    # There can be spaces in the beta releases
    release = OEGraphSimGetRelease().replace(" ", "_"),
    version = OEGraphSimGetVersion())


_is_licensed = OEGraphSimIsLicensed()

def is_licensed():
    return _is_licensed

def _get_num_bits(d):
    return d["numbits"]
    
##### Handle the atom and bond type flags for path fingerprints

# The atom and bond type flags can be specified on the command-line
# 
#   --atype=DefaultAtom --btype=BondOrder,InRing
#   --atype AtomicNumber,InRing  --btype DefaultBond,InRing
#
# The type fields may be separated by either a "," or a "|".
# The relevant OpenEye function (OEGetFPAtomType() and
# OEGetFPBondType()) use a "|" but that requires escaping for
# the shell, so I support "," as well.

# There's another conversion of the integer type values into a string
# representation, used when generating the canonical form of the
# generation parameters for the FPS output. That case uses "|"
# (and not ",") and omits the DefaultAtom and DefaultBond name.
# The result is easier to parse with the OpenEye API functions.


# Note: Version 1.0 of OEGraphSim uses different names than 2.0 (Grr!)
# OpenEye says these names will not change again. We'll see.

_atype_flags = [(OEGetFPAtomType(atype), atype) for atype in ( 
                    OEFPAtomType_Aromaticity,   # Arom
                    OEFPAtomType_AtomicNumber,  # AtmNum
                    OEFPAtomType_Chiral,        # Chiral
                    OEFPAtomType_EqHalogen,     # EqHalo
                    OEFPAtomType_FormalCharge,  # FCharge
                    OEFPAtomType_HvyDegree,     # HvyDeg
                    OEFPAtomType_Hybridization, # Hyb
                    OEFPAtomType_InRing,        # InRing
                    OEFPAtomType_EqAromatic,    # EqArom
                    OEFPAtomType_HCount,           # HCount
                    OEFPAtomType_EqHBondAcceptor,  # EqHBAcc
                    OEFPAtomType_EqHBondDonor,     # EqHBDon
                    )]
    
_btype_flags = [(OEGetFPBondType(btype), btype) for btype in
                (OEFPBondType_BondOrder,
                OEFPBondType_Chiral,
                 OEFPBondType_InRing)]

# I support the DefaultAtom and DefaultBond special values.
# (Note: complex bitflags go first; it simplifies the flag->description code)

_atype_default_flags = [("DefaultPathAtom", OEFPAtomType_DefaultPathAtom),
                        ("DefaultCircularAtom", OEFPAtomType_DefaultCircularAtom),
                        ("DefaultTreeAtom", OEFPAtomType_DefaultTreeAtom),
                        ("DefaultAtom", OEFPAtomType_DefaultAtom)]
_btype_default_flags = [("DefaultPathBond", OEFPBondType_DefaultPathBond),
                        ("DefaultCircularBond", OEFPBondType_DefaultCircularBond),
                        ("DefaultTreeBond", OEFPBondType_DefaultTreeBond),
                        ("DefaultBond", OEFPBondType_DefaultBond)]
    
_path_atype_flags = _atype_flags + _atype_default_flags + [("Default", OEFPAtomType_DefaultPathAtom)]
_path_btype_flags = _btype_flags + _btype_default_flags + [("Default", OEFPBondType_DefaultPathBond)]

_circular_atype_flags = _atype_flags + _atype_default_flags + [
    ("Default", OEFPAtomType_DefaultCircularAtom)]
_circular_btype_flags = _btype_flags + _btype_default_flags + [
    ("Default", OEFPBondType_DefaultCircularBond)]

_tree_atype_flags = _atype_flags + _atype_default_flags + [("Default", OEFPAtomType_DefaultTreeAtom)]
_tree_btype_flags = _btype_flags + _btype_default_flags + [("Default", OEFPBondType_DefaultTreeBond)]

_path_atypes = dict(_path_atype_flags)
_path_btypes = dict(_path_btype_flags)
_circular_atypes = dict(_circular_atype_flags)
_circular_btypes = dict(_circular_btype_flags)
_tree_atypes = dict(_tree_atype_flags)
_tree_btypes = dict(_tree_btype_flags)


## Go from a "," or "|" separated text field to an integer value
# Removes extra whitespace, but none should be present.

def _get_type_value(a_or_b, table, description):
    value = 0
    # Allow both "|" and "," as separators
    # (XXX OEGraphSim 2.0.0 only allows "|")
    description = description.replace("|", ",")
    for word in description.split(","):
        word = word.strip()
        try:
            value |= table[word]
        except KeyError:
            if not word:
                raise ValueError("Missing %s flag" % (a_or_b,))
            raise ValueError("Unknown %s type %r (expecting one of: %s)" % (
                a_or_b, word, ", ".join(sorted(table))))
    return value

def path_atom_description_to_value(description):
    """path_atom_description_to_value(description) -> integer

    Convert an atom description like FormalCharge,EqHalogen
    or FormalCharge|EqHalogen into its atom type value.

    This is similar to OEGetFPAtomType except both "|" and "," are
    allowed seperators and "AtomDefault" is an allowed term.
    """
    return _get_type_value("path atom", _path_atypes, description)

def path_bond_description_to_value(description):
    """bond_description_to_value(description) -> integer

    Convert an bond description like BondOrder,Chiral
    or BondOrder|Chiral into its bond type value.

    This is similar to OEGetFPBondType except both "|" and "," are
    allowed seperators and "BondDefault" is an allowed term.
    """
    return _get_type_value("path bond", _path_btypes, description)

def circular_atom_description_to_value(description):
    return _get_type_value("circular atom", _circular_atypes, description)
def circular_bond_description_to_value(description):
    return _get_type_value("circular bond", _circular_btypes, description)
def tree_atom_description_to_value(description):
    return _get_type_value("tree atom", _tree_atypes, description)
def tree_bond_description_to_value(description):
    return _get_type_value("tree bond", _tree_btypes, description)

## Go from an integer value into a canonical description

# I could use OEGetFPAtomType() and OEGetFPBondType() but I wanted
# something which has a fixed sort order even for future releases,
# which isn't part of those functions.

def _get_type_description(a_or_b, flags, value):
    words = []
    for (word, flag) in flags:
        if flag & value == flag:
            # After over 12 years of full-time use of Python,
            # I finally have a chance to use the xor operator.
            value = value ^ flag
            words.append(word)
    if value != 0:
        raise AssertionError("Unsupported %s value %d" % (a_or_b, value))
    return "|".join(words)


def path_atom_value_to_description(value):
    """atom_value_to_description(value) -> string

    Convert from an atom type string into its text description,
    separated by "|"s. The result are compatible with
    OEGetFPAtomType and are in canonical order.
    """
    return _get_type_description("path atom", _path_atype_flags, value)

def path_bond_value_to_description(value):
    """bond_value_to_description(value) -> string

    Convert from a bond type string into its text description,
    separated by "|"s. The result are compatible with
    OEGetFPBontType and are in canonical order.
    """
    return _get_type_description("path bond", _path_btype_flags, value)

def circular_atom_value_to_description(value):
    return _get_type_description("circular atom", _circular_atype_flags, value)
def circular_bond_value_to_description(value):
    return _get_type_description("circular bond", _circular_btype_flags, value)
def tree_atom_value_to_description(value):
    return _get_type_description("tree atom", _tree_atype_flags, value)
def tree_bond_value_to_description(value):
    return _get_type_description("tree bond", _tree_btype_flags, value)

    
##### Create a function which generate fingerprints

# I use functions which return functions because it was a nice way to
# hide the differences between the two fingerprinters. I also found
# that I can save a bit of time by not creating a new fingerprint each
# time. The measured speedup is about 2% for MACCS166 and 6% for path
# fingerprints.

# Just like the OEGraphMol, these fingerprints must not be reused or
# stored. They are mutated EVERY TIME. They are NOT thread-safe.
# If you need to use these in multiple threads, then make multiple
# fingerprinters.

# Believe it or not, reusing the preallocated fingerprint measurably
# helps the performance.

def get_path_fingerprinter(numbits, minbonds, maxbonds, atype, btype):
    # Extra level of error checking since I expect people will think
    # of this as part of the public API.
    if not (16 <= numbits <= 65536):
        raise ValueError("numbits must be between 16 and 65536 (inclusive)")
    if not (0 <= minbonds):
        raise ValueError("minbonds must be 0 or greater")
    if not (minbonds <= maxbonds):
        raise ValueError("maxbonds must not be smaller than minbonds")

    # XXX validate the atype and type values?
    # It's a simple mask against the | of all possible value, then test for 0.
    # However, I'm not sure what to report as the error message.
    
    fp = OEFingerPrint()
    fp.SetSize(numbits)
    data_location = int(fp.GetData())
    num_bytes = (numbits+7)//8

    def path_fingerprinter(mol):
        OEMakePathFP(fp, mol, numbits, minbonds, maxbonds, atype, btype)
        return ctypes.string_at(data_location, num_bytes)

    return path_fingerprinter

def get_maccs_fingerprinter():
    fp = OEFingerPrint()
    # Call SetSize() now to force space allocation, so I only need one GetData()
    fp.SetSize(166)
    data_location = int(fp.GetData())
    num_bytes = (166+7)//8
    
    def maccs_fingerprinter(mol):
        OEMakeMACCS166FP(fp, mol)
        return ctypes.string_at(data_location, num_bytes)

    return maccs_fingerprinter

def get_circular_fingerprinter(numbits, minradius, maxradius, atype, btype):
    # Extra level of error checking since I expect people will think
    # of this as part of the public API.
    if not (16 <= numbits <= 65536):
        raise ValueError("numbits must be between 16 and 65536 (inclusive)")
    if not (0 <= minradius):
        raise ValueError("minradius must be 0 or greater")
    if not (minradius <= maxradius):
        raise ValueError("maxradius must not be smaller than minradius")

    fp = OEFingerPrint()
    fp.SetSize(numbits)
    data_location = int(fp.GetData())
    num_bytes = (numbits+7)//8

    def circular_fingerprinter(mol):
        OEMakeCircularFP(fp, mol, numbits, minradius, maxradius, atype, btype)
        return ctypes.string_at(data_location, num_bytes)

    return circular_fingerprinter


def get_tree_fingerprinter(numbits, minbonds, maxbonds, atype, btype):
    # Extra level of error checking since I expect people will think
    # of this as part of the public API.
    if not (16 <= numbits <= 65536):
        raise ValueError("numbits must be between 16 and 65536 (inclusive)")
    if not (0 <= minbonds):
        raise ValueError("minbonds must be 0 or greater")
    if not (minbonds <= maxbonds):
        raise ValueError("maxbonds must not be smaller than minbonds")

    fp = OEFingerPrint()
    fp.SetSize(numbits)
    data_location = int(fp.GetData())
    num_bytes = (numbits+7)//8

    def tree_fingerprinter(mol):
        OEMakeTreeFP(fp, mol, numbits, minbonds, maxbonds, atype, btype)
        return ctypes.string_at(data_location, num_bytes)

    return tree_fingerprinter
        

### A note on fingerprints and ctypes.string_at

# The FPS format and OEFingerPrint.GetData() values used identical bit
# and byte order. Bytes are in little-endian order and bits are in
# big-endian order. That means I can use GetData() to get the
# underlying C storage area, use ctypes to turn that into a Python
# string, which I then hex encode.

# The other option is to use OEFingerPrint.ToHexString(). But that's
# pure little endian, so I would need a transposition to make the bits
# be what I want them to be. OEChem's hex strings also end with a flag
# which says how many extra bits to trim, which I don't need since I
# handle it a different way.

# Here's some info about the bit order, which I tested by setting a
# few bits though the API then seeing what changed in the hex string
# and in the underlying GetData() field.

# The bit pattern
#   01234567 89ABCDEF  pure little endian
#   10011100 01000011
#   
#         93 2C    using ToHexString()  (pure little endian)
#       0x39 c2    using hex(ord(GetData())) (litle endian byte, big endian bit)
#
#   76543210 FEDCBA98
#   00111001 11000010   little endian byte, big endian bit


################ Handle formats

# Make format names to OEChem format types
_formats = {
    "smi": OEFormat_SMI,
    "ism": OEFormat_ISM,  # old format, don't use
    "usm": OEFormat_USM,
    "can": OEFormat_CAN,

    "sdf": OEFormat_SDF,
    "mol": OEFormat_SDF,

    "skc": OEFormat_SKC,
    "mol2": OEFormat_MOL2,
    "mmod": OEFormat_MMOD,
    "oeb": OEFormat_OEB,
}

# Some trickiness to verify that the format specification is
# supported, but without doing anything (yet) to set those flags.

# I return a function which will set the file stream parameters
# correctly.

def _do_nothing(ifs):
    pass

# Return a function which can be use to set the ifs format, flavor, and compression properties
    
def get_format_setter(format, compression):
    if compression:
        if compression == "gz":
            is_compressed = True
        else:
            raise ValueError("OEChem does not support compression type %r" % (compression,))
    else:
        is_compressed = False

    oeformat = _formats.get(format.lower(), None)
    if oeformat is None:
        raise ValueError("OEChem does not support the %r format" % (format,))
        
    def set_format(ifs):
        ifs.SetFormat(oeformat)
        # OEChem 20170208 has a bug where Setgz(True) on a file where
        # Getgz()==True causes data loss and Setgz(False) causes a
        # segmentation fault. I can work around the first case by
        # not setting 'gz' when 'gz' is already set. I can't work
        # around the second problem. I classify it as a vendor bug,
        # and let the segmentation fault percolate through.
        # Reported to OpenEye on 2017-04-25 in an email titled
        #   'can't use Setgz() to change the gz state of an oemolostream'
        if is_compressed and ifs.Getgz():
            pass
        else:
            ifs.Setgz(is_compressed)
        
    return set_format

def _open_stdin(set_format, aromaticity_flavor):
    ifs = oemolistream()
    ifs.open()
    set_format(ifs)
    
    if aromaticity_flavor is not None:
        flavor = ifs.GetFlavor(ifs.GetFormat())
        flavor |= aromaticity_flavor
        ifs.SetFlavor(ifs.GetFormat(), flavor)

    return ifs

def _open_ifs(filename, set_format, aromaticity_flavor):
    ifs = oemolistream()

    if not ifs.open(filename):
        # Let Python try to do better error reporting.
        open(filename).close()
        # If that didn't work, give up and fake it.
        # (Did manual coverage testing for this. The test cases I can
        # think of, like tricky timing, are too tricky.)
        raise IOError(errno.EIO, "OEChem cannot open the file", filename)

    set_format(ifs)

    if aromaticity_flavor is not None:
        flavor = ifs.GetFlavor(ifs.GetFormat())
        flavor |= aromaticity_flavor
        ifs.SetFlavor(ifs.GetFormat(), flavor)
    
    return ifs

# This code is a bit strange. It needs to do eager error checking but
# lazy parsing. That is, it needs to check right away that the file
# can be opened (if it exists) and the format is understood. But it
# can wait until later to actually parse the files.

_aromaticity_sorted = (
#    ("default", None),
    ("openeye", OEIFlavor_Generic_OEAroModelOpenEye),
    ("daylight", OEIFlavor_Generic_OEAroModelDaylight),
    ("tripos", OEIFlavor_Generic_OEAroModelTripos),
    ("mdl", OEIFlavor_Generic_OEAroModelMDL),
    ("mmff", OEIFlavor_Generic_OEAroModelMMFF),
    )
_aromaticity_flavors = dict(_aromaticity_sorted)
_aromaticity_flavor_names = [pair[0] for pair in _aromaticity_sorted]
_aromaticity_flavors[None] = OEIFlavor_Generic_OEAroModelOpenEye
del _aromaticity_sorted, pair

# If unspecified, use "openeye" (this is what OEChem does internally)
_aromaticity_flavors[None] = _aromaticity_flavors["openeye"]   # Allow "None"

def is_valid_format(filename, format):
    format_name, compression = io.normalize_input_format(filename, format,
                                                         default=("smi", ""))
    if compression not in ("", "gz"):
        return False
    try:
        get_format_setter(format_name, compression)
        return True
    except ValueError:
        return False

def is_valid_aromaticity(aromaticity):
    return aromaticity in _aromaticity_flavors

# Part of the code (parameter checking, opening the file) are eager.
# Actually reading the structures is lazy.

def read_structures(source, format=None, id_tag=None, reader_args=None, errors="strict", location=None):
    if reader_args is None:
        aromaticity = None
    else:
        aromaticity = reader_args.get("aromaticity", None)
    try:
        aromaticity_flavor = _aromaticity_flavors[aromaticity]
    except KeyError:
        raise ValueError("Unsupported aromaticity model %r" % (aromaticity,))
    error_handler = io.get_parse_error_handler(errors)

    # Check that that the format is known
    format_name, compression = io.normalize_input_format(source, format,
                                                         default=("smi", ""))

    if compression not in ("", "gz"):
        raise ValueError("OEChem does not support compression type %r for %r" % (compression, source))
    if location is None:
        location = io.Location.from_source(source)
    location.save(record_format=format_name)

    set_format = get_format_setter(format_name, compression)

    # Input is from a file
    if source is None:
        ifs = _open_stdin(set_format, aromaticity_flavor)
    else:
        ifs = _open_ifs(source, set_format, aromaticity_flavor)

    # Only SD files can take the id_tag
    if ifs.GetFormat() != OEFormat_SDF:
        id_tag = None

    # Lazy structure reader
    return _iter_structures(ifs, id_tag, error_handler, location)

def _iter_structures(ifs, id_tag, error_handler, location):
    recno = 0
    mol = None

    def get_recno():
        return recno
    def get_mol():
        return mol
    location.register(get_recno = get_recno,
                      get_mol = get_mol)
    try:
        if id_tag is None:
            for recno, mol in enumerate(ifs.GetOEGraphMols(), 1):
                title = mol.GetTitle()
                yield title, mol
        else:
            for recno, mol in enumerate(ifs.GetOEGraphMols(), 1):
                id = OEGetSDData(mol, id_tag)
                if not id:
                    # Either it's empty or missing
                    if not OEHasSDData(mol, id_tag):
                        # Missing.
                        id = None
                yield id, mol
    finally:
        ifs.close()
        location.save(recno = get_recno(),
                      mol = None)


def _read_fingerprints(structure_reader, fingerprinter):
    for (id, mol) in structure_reader:
        yield id, fingerprinter(mol)

from .types import FingerprintFamilyConfig, nonnegative_int

def _read_structures(metadata, source, format, id_tag, reader_args, errors, location):
    return read_structures(source, format, id_tag=id_tag, reader_args=reader_args,
                           errors=errors, location=location)

def _correct_numbits(s):
    try:
        if not s.isdigit():
            raise ValueError
        i = int(s)
        if not (16 <= i <= 65536):
            raise ValueError
    except ValueError:
        raise ValueError("must be between 16 and 65536 bits")
    return i

_base = FingerprintFamilyConfig(
    software = SOFTWARE + " " + CHEMFP_SOFTWARE,
    read_structures = _read_structures)


######### These are appropriate for OEGraphSim 2.0  #############

_path_ff = OpenEyePathFingerprintFamily_v2 = _base.clone(
    name = "OpenEye-Path/2",
    format_string = ("numbits=%(numbits)s minbonds=%(minbonds)s "
                     "maxbonds=%(maxbonds)s atype=%(atype)s btype=%(btype)s"),
    make_fingerprinter = get_path_fingerprinter,
    num_bits = _get_num_bits,
    )
_path_ff.add_argument("numbits", decoder=_correct_numbits, metavar="INT", default=4096,
                      help="number of bits in the path fingerprint")

_path_ff.add_argument("minbonds", decoder=nonnegative_int, metavar="INT", default=0,
                      help="minimum number of bonds in the path fingerprint")

_path_ff.add_argument("maxbonds", decoder=nonnegative_int, metavar="INT", default=5,
                      help="maximum number of bonds in the path fingerprint")

_path_ff.add_argument("atype", decoder=path_atom_description_to_value,
                      encoder=path_atom_value_to_description,
                      help="atom type", default=OEFPAtomType_DefaultPathAtom)
_path_ff.add_argument("btype", decoder=path_bond_description_to_value,
                      encoder=path_bond_value_to_description,
                      help="bond type", default=OEFPBondType_DefaultPathBond)


if _is_licensed:
    _PROBE_MOL = OEGraphMol()
    _PROBE_FP = OEFingerPrint()
    _PROBE_MOL.NewAtom(2)
    OEMakeMACCS166FP(_PROBE_FP, _PROBE_MOL)
    _maccs_version = {
        "2.0.0": "2",
        "2.2.0": "3",
        }.get(_PROBE_FP.GetFPTypeBase().GetFPVersionString(), "unknown")
else:
    _maccs_version = "unknown"

def _check_maccs_version(version):
    def make_fingerprinter(*args, **kwargs):
        if _maccs_version != version:
            raise TypeError("This version of OEChem does not support the RDKit-MACCS/%s fingerprint" % (version,))
        return get_maccs_fingerprinter(*args, **kwargs)
    return make_fingerprinter

    
OpenEyeMACCSFingerprintFamily_v2 = _base.clone(
    name = "OpenEye-MACCS166/2",
    make_fingerprinter = _check_maccs_version("2"),
    num_bits = 166,
    )

OpenEyeMACCSFingerprintFamily_v3 = _base.clone(
    name = "OpenEye-MACCS166/3",
    make_fingerprinter = _check_maccs_version("3"),
    num_bits = 166,
    )

_circular_ff = OpenEyeCircularFingerprintFamily_v2 = _base.clone(
    name = "OpenEye-Circular/2",
    format_string = ("numbits=%(numbits)s minradius=%(minradius)s "
                     "maxradius=%(maxradius)s atype=%(atype)s btype=%(btype)s"),
    num_bits = _get_num_bits,
    make_fingerprinter = get_circular_fingerprinter)
_circular_ff.add_argument("numbits", decoder=_correct_numbits, metavar="INT", default=4096,
                          help="number of bits in the circular fingerprint")
_circular_ff.add_argument("minradius", decoder=nonnegative_int, metavar="INT", default=0,
                          help="minimum radius for the circular fingerprint")
_circular_ff.add_argument("maxradius", decoder=nonnegative_int, metavar="INT", default=5,
                          help="maximum radius for the circular fingerprint")
_circular_ff.add_argument("atype", decoder=circular_atom_description_to_value,
                          encoder=circular_atom_value_to_description,
                          help="atom type", default=OEFPAtomType_DefaultCircularAtom)
_circular_ff.add_argument("btype", decoder=circular_bond_description_to_value,
                 encoder=circular_bond_value_to_description,
                 help="bond type", default=OEFPBondType_DefaultCircularBond)

_tree_ff = OpenEyeTreeFingerprintFamily_v2 = _base.clone(
    name = "OpenEye-Tree/2",
    format_string = ("numbits=%(numbits)s minbonds=%(minbonds)s "
                     "maxbonds=%(maxbonds)s atype=%(atype)s btype=%(btype)s"),
    num_bits = _get_num_bits,
    make_fingerprinter = get_tree_fingerprinter)
_tree_ff.add_argument("numbits", decoder=_correct_numbits, metavar="INT", default=4096,
                      help="number of bits in the tree fingerprint")
_tree_ff.add_argument("minbonds", decoder=nonnegative_int, metavar="INT", default=0,
                      help="minimum number of bonds in the tree fingerprint")
_tree_ff.add_argument("maxbonds", decoder=nonnegative_int, metavar="INT", default=4,
                      help="maximum number of bonds in the tree fingerprint")
_tree_ff.add_argument("atype", decoder=tree_atom_description_to_value,
                      encoder=tree_atom_value_to_description,
                      help="atom type", default=OEFPAtomType_DefaultTreeAtom)
_tree_ff.add_argument("btype", decoder=tree_bond_description_to_value,
                      encoder=tree_bond_value_to_description,
                      help="bond type", default=OEFPBondType_DefaultTreeBond)




import sys
import unittest2
import doctest

def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite("chemfp.encodings"))
    return tests

if __name__ == "__main__":
    unittest2.main()



# Copyright (c) 2010-2017 Andrew Dalke Scientific, AB (Sweden)
# See the contents of "chemfp/__init__.py" for full license details.

from __future__ import absolute_import

from openbabel import OBSmartsPattern, OBBitVec

from . import openbabel
from . import pattern_fingerprinter
from . import types
from . import __version__ as chemfp_version

SOFTWARE = openbabel.SOFTWARE + (" chemfp/%s" % (chemfp_version,))

# HasMatch was added to OpenBabel 2.3 .
_HAS_HASMATCH = hasattr(OBSmartsPattern(), "HasMatch")

class Matcher(object):
    def __init__(self, ob_matcher):
        self.ob_matcher = ob_matcher

    if _HAS_HASMATCH:
        def HasMatch(self, mol):
            return self.ob_matcher.HasMatch(mol)
    else:
        def HasMatch(self, mol):
            return self.ob_matcher.Match(mol, True)
        
    def NumUniqueMatches(self, mol):
        self.ob_matcher.Match(mol)
        return sum(1 for x in self.ob_matcher.GetUMapList())

class HydrogenMatcher(object):
    def __init__(self, max_count):
        self.max_count = max_count
    def HasMatch(self, mol):
        for i in range(1, mol.NumAtoms()+1):
            atom = mol.GetAtom(i)
            if atom.GetAtomicNum() == 1:
                return True
            if atom.ImplicitHydrogenCount():
                return True
        return False
        
    def NumUniqueMatches(self, mol):
        count = 0
        for i in range(1, mol.NumAtoms()+1):
            atom = mol.GetAtom(i)
            if atom.GetAtomicNum() == 1:
                count += 1
            count += atom.ImplicitHydrogenCount()
            if count >= self.max_count:
                return count
        return count

class AromaticRings(object):
    def __init__(self, max_count):
        self.max_count = max_count
        # The single ring case is easy; if there's an aromatic atom in a ring
        # then there's an aromatic ring.
        self._single_matcher = OBSmartsPattern()
        assert self._single_matcher.Init("[aR]")
        
    if _HAS_HASMATCH:
        def HasMatch(self, mol):
            return self._single_matcher.HasMatch(mol)
    else:
        def HasMatch(self, mol):
            return self._single_matcher.Match(mol, True)
        

    def NumUniqueMatches(self, mol):
        num_rings = 0
        for ring in mol.GetSSSR():
            if ring.IsAromatic():
                num_rings += 1
                if num_rings == self.max_count:
                    break
        return num_rings


###

class HeteroAromaticRings(object):
    def __init__(self, max_count):
        self.max_count = max_count
        # The single ring case is easy; if there's an non-carbon aromatic atom
        # in a ring then there's a hetero-aromatic ring
        self._single_matcher = OBSmartsPattern()
        assert self._single_matcher.Init("[aR;!#6]")

    if _HAS_HASMATCH:
        def HasMatch(self, mol):
            return self._single_matcher.HasMatch(mol)
    else:
        def HasMatch(self, mol):
            return self._single_matcher.Match(mol, True)
        

    def NumUniqueMatches(self, mol):
        num_rings = 0
        for ring in mol.GetSSSR():
            if (ring.IsAromatic() and
                  any(mol.GetAtom(atom_id).GetAtomicNum() != 6 for atom_id in ring._path)):
                num_rings += 1
                if num_rings == self.max_count:
                    break
        return num_rings


class NumFragments(object):
    def __init__(self, max_count):
        if max_count >= 3:
            raise NotImplemented("<fragment> >= 3 not implemented")
    def HasMatch(self, mol):
        # Has at least one fragment, which is the same as having atoms
        return (mol.NumAtoms() > 0)
    def NumUniqueMatches(self, mol):
        # If the largest fragment is not the same size as the entire molecule
        # then there are at least two fragments
        n = mol.NumAtoms()
        if n == 0:
            # No atoms means no fragments
            return 0
        
        v = OBBitVec()
        mol.FindLargestFragment(v)
        if v.CountBits() == n:
            return 1
        
        return 2

_pattern_classes = {
    "<H>": HydrogenMatcher,
    "<aromatic-rings>": AromaticRings,
    "<hetero-aromatic-rings>": HeteroAromaticRings,
    "<fragments>": NumFragments,
    }

def ob_compile_pattern(pattern, max_count):
    if pattern in _pattern_classes:
        return _pattern_classes[pattern](max_count)
    
    if pattern.startswith("<"):
        raise NotImplementedError(pattern)

    matcher = OBSmartsPattern()
    if not matcher.Init(pattern):
        raise pattern_fingerprinter.UnsupportedPatternError(
            pattern, "Uninterpretable SMARTS pattern")
    return Matcher(matcher)

class OBPatternFingerprinter(pattern_fingerprinter.PatternFingerprinter):
    def __init__(self, patterns):
        assert patterns is not None
        super(OBPatternFingerprinter, self).__init__(patterns, ob_compile_pattern)
        
    def fingerprint(self, mol):
        bytes = [0] * self.num_bytes
        for matcher, largest_count, count_info_tuple in self.matcher_definitions:
            #print matcher, largest_count, count_info_tuple
            if largest_count == 1:
                if matcher.HasMatch(mol):
                    count_info = count_info_tuple[0]
                    bytes[count_info.byteno] |= count_info.bitmask
            else:
                # There's no good way to get the number of unique
                # matches without iterating over all of them.
                actual_count = matcher.NumUniqueMatches(mol)
                for count_info in count_info_tuple:
                    if actual_count >= count_info.count:
                        bytes[count_info.byteno] |= count_info.bitmask
                    else:
                        break
        return "".join(map(chr, bytes))


class _CachedFingerprinters(dict):
    def __missing__(self, name):
        patterns = pattern_fingerprinter._load_named_patterns(name)
        fingerprinter = OBPatternFingerprinter(patterns)
        self[name] = fingerprinter
        return fingerprinter
_cached_fingerprinters = _CachedFingerprinters()

# XXX Why are there two "Fingerprinter" classes?
# XX Shouldn't they be merged?

_base = openbabel._base.clone(
    software = SOFTWARE)
    

SubstructOpenBabelFingerprinter_v1 = _base.clone(
    name = "ChemFP-Substruct-OpenBabel/1",
    num_bits = 881,
    make_fingerprinter = lambda: _cached_fingerprinters["substruct"].fingerprint)
    

RDMACCSOpenBabelFingerprinter_v1 = _base.clone(
    name = "RDMACCS-OpenBabel/1",
    num_bits = 166,
    make_fingerprinter = lambda: _cached_fingerprinters["rdmaccs"].fingerprint)

RDMACCSOpenBabelFingerprinter_v2 = _base.clone(
    name = "RDMACCS-OpenBabel/2",
    num_bits = 166,
    make_fingerprinter = lambda: _cached_fingerprinters["rdmaccs2"].fingerprint)
